from functools import wraps
from typing import Sequence


def _no_list(v):
    return tuple(v) if isinstance(v, Sequence) and not isinstance(v, str) else v


def list_to_tuple(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        args = (_no_list(arg) for arg in args)
        kwargs = {k: _no_list(v) for k, v in kwargs.items()}
        return func(*args, **kwargs)

    return wrapper



from kanon.tables.htable_reader import read_historical
from kanon.units import Historical


def test_read_historical():

    assert read_historical(["12", "0"], 0) == Historical(1, 0, 0)
    assert read_historical(["11", "0"], 0) == Historical(11, 0)
    assert read_historical(["5"], 0) == Historical(5)
    assert read_historical(["9", "10", "5"], 0) == Historical(9, 10, 5)
    assert read_historical(["9", "10", "5"], 1) == Historical((9, 10), (5,))



"""Configure Test Suite.

This file is used to configure the behavior of pytest when using the Astropy
test infrastructure. It needs to live inside the package in order for it to
get picked up when running the tests inside an interpreter using
packagename.test

"""

from hypothesis.strategies._internal.core import tuples


def _hypothesis_sexagesimal_strategy():
    """We define hypothesis strategy to generate Sexagesimal values in tests"""
    from hypothesis import HealthCheck, settings
    from hypothesis.strategies import (
        builds,
        decimals,
        integers,
        lists,
        register_type_strategy,
        sampled_from,
    )

    from kanon.units import Historical, Sexagesimal

    settings.register_profile("def", suppress_health_check=(HealthCheck.too_slow,))
    settings.load_profile("def")

    strat = builds(
        Sexagesimal,
        lists(integers(0, 59), max_size=2),
        lists(integers(0, 59), max_size=2),
        remainder=decimals(0, 1).filter(lambda x: x != 1),
        sign=sampled_from((-1, 1)),
    )
    register_type_strategy(Sexagesimal, strat)

    strat = builds(
        Historical,
        tuples(integers(0, 9), integers(0, 11), integers(0, 29)),
        lists(integers(0, 59), max_size=2),
        remainder=decimals(0, 1).filter(lambda x: x != 1),
        sign=sampled_from((-1, 1)),
    )
    register_type_strategy(Historical, strat)


# Uncomment the last two lines in this block to treat all DeprecationWarnings as
# exceptions. For Astropy v2.0 or later, there are 2 additional keywords,
# as follow (although default should work for most cases).
# To ignore some packages that produce deprecation warnings on import
# (in addition to 'compiler', 'scipy', 'pygments', 'ipykernel', and
# 'setuptools'), add:
#     modules_to_ignore_on_import=['module_1', 'module_2']
# To ignore some specific deprecation warning messages for Python version
# MAJOR.MINOR or later, add:
#     warnings_to_ignore_by_pyver={(MAJOR, MINOR): ['Message to ignore']}
# from astropy.tests.helper import enable_deprecations_as_exceptions  # noqa
# enable_deprecations_as_exceptions()



import pandas as pd
import pytest

from kanon.tables.symmetries import (
    OutOfBoundsOriginError,
    OverlappingSymmetryError,
    Symmetry,
)


class TestSymmetry:
    def test_init(self):
        with pytest.raises(ValueError):
            Symmetry("a")

    def test_symmetry_call(self):
        symmirror = Symmetry("mirror", sign=-1)
        symperiod = Symmetry("periodic")

        df = pd.DataFrame()

        assert symmirror(df).equals(df)

        df = pd.DataFrame({"a": [1, 2, 4], "b": [4, 5, 6]}).set_index("a")

        symdf = symmirror(df)

        assert len(symdf) == 2 * len(df) - 1

        assert list(symdf.index) == [1, 2, 4, 6, 7]

        symdf = symperiod(df)

        assert len(symdf) == 2 * len(df)

        with pytest.raises(ValueError):
            Symmetry("mirror", source=(2, 1))

        sym = Symmetry("periodic", source=(0, 1))
        with pytest.raises(OutOfBoundsOriginError):
            sym(df)

        sym = Symmetry("periodic", targets=[1])
        with pytest.raises(OverlappingSymmetryError):
            sym(df)

        sym = Symmetry("periodic", source=(1, 2), targets=[5])
        assert sym(df).index[-1] == 6

    def test_mirror_multi_targets(self):
        sym = Symmetry("mirror", targets=[10, 15])
        df = pd.DataFrame({"a": [2, 3, 5], "b": [7, 6, 9]}).set_index("a")

        res = df.pipe(sym)

        assert len(res) == 9

        assert list(res.index) == [2, 3, 5, 10, 12, 13, 15, 17, 18]
        assert list(res["b"]) == [7, 6, 9, 9, 6, 7, 9, 6, 7]



from math import isclose
from typing import Tuple

import astropy.units as u
import hypothesis.strategies as st
import numpy as np
import pandas as pd
import pytest
from astropy.table import setdiff
from astropy.table.operations import join
from astropy.units.core import UnitConversionError
from hypothesis import assume
from hypothesis.core import given

from kanon.tables.htable import HTable, join_multiple
from kanon.tables.symmetries import Symmetry

sample = {"a": [1, 2, 3, 4, 5], "b": [5, 9, 12, 25, 26]}


def make_sample_table(index="a"):
    return HTable(sample, index=index)


def test_init():

    table = HTable(sample)
    assert table[2]["b"] == 12

    with pytest.raises(IndexError):
        table.get(3)

    table.add_index("b")
    assert table.get(5) == 1


gen_table_strategy = st.builds(
    HTable,
    st.lists(
        st.tuples(
            st.integers(min_value=int(-1e15), max_value=int(1e15)),
            st.floats(allow_nan=False, allow_infinity=False, width=16),
        ),
        min_size=2,
        unique_by=(lambda x: x[0]),
    ).map(lambda x: list(zip(*sorted(x)))),
    names=st.just(("A", "B")),
    index=st.just("A"),
)


@given(gen_table_strategy)
def test_symmetry(tab: HTable):
    assume(len(tab) > 1)
    tab.symmetry = [Symmetry("mirror", sign=-1)]
    df = tab.to_pandas()
    assert df["B"].iloc[0] == -df["B"].iloc[-1]
    tab.symmetry = [Symmetry("periodic")]
    df = tab.to_pandas()
    assert df["B"].iloc[0] == df["B"].iloc[len(tab)]


@given(
    gen_table_strategy.flatmap(
        lambda x: st.tuples(
            st.floats(
                min_value=float(min(x["A"])),
                max_value=float(max(x["A"])),
                allow_nan=False,
            ),
            st.just(x),
        )
    )
)
def test_interpolation(hypo: Tuple[float, HTable]):
    key, tab = hypo
    fres = np.interp(
        key, [float(x) for x in tab.columns[0]], [float(x) for x in tab.columns[1]]
    )
    sres = tab.get(key)
    assert isclose(fres, float(sres), abs_tol=1e9)


def test_apply():
    tab = make_sample_table()

    assert np.array_equal(
        tab.apply("b", lambda x: x + 2)["b"],
        np.add(tab["b"], np.array([2] * len(tab))),
    )

    tab_float = tab.apply("b", lambda x: x + 0.3)
    assert tab_float["b"].dtype == np.dtype("float64")
    tab_int = tab_float.apply("b", round, "rounded")
    assert tab_int["rounded"].dtype == np.dtype("int64")

    def tostr(x):
        return f"{x}a"

    tab_str = tab.apply("b", tostr)
    assert tab_str["b"].dtype.char == "U"


def test_index():
    tab = make_sample_table()

    assert tab.loc[1]["b"] == 5
    with pytest.raises(KeyError):
        tab.loc[6]

    new_tab = tab.copy(set_index="b")
    assert new_tab.loc[5]["a"] == 1
    with pytest.raises(KeyError):
        new_tab.loc[1]


def test_populate():
    tab = make_sample_table(index="b")

    pop_array = [i for i in range(5, 15)]

    populated = tab.populate(pop_array, method="interpolate")

    assert len(populated) == len(set(populated["b"]).union(pop_array))

    assert populated.loc[6]["a"] == tab.get(6)


def test_diff():
    tab = make_sample_table()

    assert list(tab.diff()["b"]) == [0, 4, 3, 13, 1]
    assert list(tab.diff(prepend=[3])["b"]) == [2, 4, 3, 13, 1]
    assert list(tab.diff(append=[24])["b"]) == [4, 3, 13, 1, -2]

    with pytest.raises(ValueError):
        tab.diff(prepend=[1, 1])


def test_fill():
    tab = make_sample_table()
    tab["b"] = tab["b"].astype("float64")

    tab = tab.populate([i / 4 for i in range(4, 16)])

    tab_filled = tab.fill("distributed_convex", (2, 4))
    assert list(tab_filled.loc[3:4]["b"]) == [12, 15, 18, 21, 25]

    tab_filled = tab.fill("distributed_concave", (2, 4))
    assert list(tab_filled.loc[3:4]["b"]) == [12, 16, 19, 22, 25]

    tab_unmasked = tab.filled(50)

    assert (
        len(setdiff(tab_unmasked, tab.fill("distributed_convex", (4, 5)).filled(50)))
        == 0
    )
    assert (
        len(
            setdiff(tab_unmasked, tab.fill("distributed_convex", (3.5, 3.5)).filled(50))
        )
        == 0
    )

    def fill_50(df: pd.DataFrame):
        return df.fillna(50)

    assert len(setdiff(tab_unmasked, tab.fill(fill_50))) == 0

    with pytest.raises(ValueError) as err:
        tab.fill("distributed_convex", (2.5, 4))
    assert str(err.value) == "First and last rows must not be masked"

    with pytest.raises(ValueError) as err:
        tab.fill("not a method", (2, 4))
    assert str(err.value) == "Incorrect fill method"


def test_join_multiple():
    data = [HTable({"a": x, "b": x}) for x in [[1, 2, 3], [8, 10, 13], [435, 13, 3.5]]]
    multiple_joined = join_multiple(*data, join_type="outer")
    single_joined = join(
        join(data[0], data[1], join_type="outer"), data[2], join_type="outer"
    )

    assert len(setdiff(multiple_joined, single_joined)) == 0


def test_plot():
    tab = make_sample_table()
    plot = tab.plot2d(".")
    assert len(plot) == 1
    assert plot[0].get_marker() == "."


def test_get_with_quantity():
    tab = make_sample_table()

    tab["a"].unit = u.degree

    assert tab.get(3) == 12
    assert tab.get(3 * u.degree) == 12

    assert tab.get(180 * u.arcmin) == 12

    with pytest.raises(UnitConversionError):
        assert tab.get(3 * u.deg_C)



from .radices import BasedReal, RadixBase, radix_registry

__all__ = ["RadixBase", "BasedReal"]

# Load all common radices

Sexagesimal = radix_registry["Sexagesimal"]
Historical = radix_registry["Historical"]
HistoricalDecimal = radix_registry["HistoricalDecimal"]
IntegerAndSexagesimal = radix_registry["IntegerAndSexagesimal"]
Temporal = radix_registry["Temporal"]



import math
from math import isclose
from typing import Tuple

import astropy.units as u
import hypothesis.strategies as st
import numpy as np
import pytest
from astropy.units.quantity import Quantity
from hypothesis.core import given

from kanon.tables import HTable
from kanon.units import Sexagesimal


def test_read():
    table: HTable = HTable.read("180", format="dishas")

    assert table["Mean Argument of the Sun"].unit is u.degree
    assert table["Mean Argument of the Sun"].basedtype is Sexagesimal
    assert table["Entries"].significant == 2

    assert table.values_equal(HTable.read(180))

    assert table.loc[Sexagesimal(1)] == table[0]

    assert table.loc[Sexagesimal(3)]["Entries"].equals(Sexagesimal("-00 ; 06,27"))

    assert table.loc[Sexagesimal(3)]["Entries"].equals(
        table.get(Sexagesimal(3), with_unit=False)
    )

    assert len(table.symmetry) == 1
    sym = table.symmetry[0]

    assert table.get(1) == -table.get(37)

    assert sym.symtype == "mirror"

    assert "Sexagesimal" in repr(table)

    with pytest.raises(FileNotFoundError):
        HTable.read(181, format="dishas")


def test_read_double():
    table = HTable.read(287)
    assert len(table) == 3
    sub_table = table.get(2)
    assert len(sub_table) == 15


gen_table_strategy = st.builds(
    HTable,
    st.lists(
        st.tuples(
            st.from_type(Sexagesimal).map(lambda x: x.resize(2).truncate()),
            st.from_type(Sexagesimal).map(lambda x: x.resize(2).truncate()),
        ),
        min_size=2,
        max_size=7,
        unique_by=(lambda x: x[0]),
    ).map(lambda x: list(zip(*sorted(x)))),
    names=st.just(("A", "B")),
    index=st.just("A"),
    dtype=st.just([object, object]),
)


@given(
    gen_table_strategy.flatmap(
        lambda x: st.tuples(
            st.floats(
                min_value=float(min(x["A"])),
                max_value=float(max(x["A"])),
                allow_nan=False,
            ),
            st.just(x),
        )
    )
)
def test_interpolation(hypo: Tuple[float, HTable]):
    key, tab = hypo
    fres = np.interp(
        key, [float(x) for x in tab.columns[0]], [float(x) for x in tab.columns[1]]
    )
    sres = tab.get(key)
    assert isclose(fres, float(sres), abs_tol=1e9)


@given(gen_table_strategy)
def test_quantity(tab: HTable):
    tab["A"].unit = u.degree
    tab["B"].unit = u.degree

    value = tab.get(tab["A"][0])

    assert isinstance(value, Quantity)
    assert value.unit is u.degree
    assert isinstance(value.value, Sexagesimal)


@given(gen_table_strategy)
def test_loc_slice(tab: HTable):
    sliced = tab.loc[tab["A"][1] :]
    assert not isinstance(sliced, HTable) or len(sliced) == len(tab) - 1


sin_table = HTable(
    [
        list(Sexagesimal.range(91)),
        [
            round(Sexagesimal.from_float(math.sin(x * math.pi / 180), 3))
            for x in range(91)
        ],
    ],
    names=("Arg", "Val"),
    index="Arg",
)
sin_table_grid = sin_table[
    [i for i in range(91) if i <= 60 and i % 12 == 0 or i > 60 and i % 6 == 0]
]


def test_based_fill():
    sin_table_pop_euclidean = sin_table_grid.populate(list(Sexagesimal.range(91)))
    sin_table_pop_euclidean.fill("distributed_convex")
    assert len(sin_table_pop_euclidean) == 91


def test_based_populate():
    tab = sin_table_grid.populate(list(range(91)))
    assert tab["Arg"].basedtype == Sexagesimal
    assert len(tab) == 91

    tab_int = sin_table_grid.copy()
    tab_int.remove_indices("Arg")
    tab_int["Arg"] = tab_int["Arg"].astype(int)
    tab_int.set_index("Arg")

    tab = tab_int.populate(list(Sexagesimal.range(91)))
    assert tab["Arg"].basedtype is None
    assert tab["Arg"].dtype == int
    assert len(tab) == 91



import importlib
from functools import wraps
from typing import Callable, Optional, Type, cast

import numpy as np
from astropy.table import Column
from astropy.table.column import ColumnInfo

from kanon.units import BasedReal
from kanon.units.precision import Truncable

__all__ = ["HColumn", "HColumnInfo"]


def _patch_dtype_info_name(func: Callable, col_arg: int):
    """
    Wrapper monkey patching `dtype_info_name` to replace it with a column `BasedReal`
    name if possible.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        col = args[col_arg]
        if col.basedtype:
            module = importlib.import_module(func.__module__)
            base_info_name = module.dtype_info_name
            module.dtype_info_name = lambda _: col.basedtype.__name__
            res = func(*args, **kwargs)
            module.dtype_info_name = base_info_name
            return res
        return func(*args, **kwargs)

    return wrapper


class HColumnInfo(ColumnInfo):
    """
    `~ColumnInfo` with `basedtype`
    """

    attrs_from_parent = ColumnInfo.attrs_from_parent | {"basedtype"}
    attr_names = ColumnInfo.attr_names | {"basedtype"}

    def new_like(self, cols, length, metadata_conflicts="warn", name=None):
        attrs = self.merge_cols_attributes(
            cols,
            metadata_conflicts,
            name,
            ("meta", "unit", "format", "basedtype", "description"),
        )

        return self._parent_cls(length=length, **attrs)


class HColumn(Column, Truncable):
    """
    `~astropy.table.Column` subclass with better support of `~kanon.units.radices.BasedReal`
    values.
    """

    info = HColumnInfo()

    _basedtype: Optional[Type[BasedReal]] = None

    def __new__(
        cls,
        data=None,
        name=None,
        dtype=None,
        shape=(),
        length=0,
        description=None,
        unit=None,
        format=None,
        meta=None,
        copy=False,
        copy_indices=True,
        basedtype: Optional[Type[BasedReal]] = None,
    ):

        if data is None and basedtype:
            data = np.zeros((length,) + shape, dtype="O")
            data = np.vectorize(basedtype.from_int)(data)

        self = super().__new__(
            cls,
            data=data,
            name=name,
            dtype=dtype,
            shape=shape,
            length=length,
            description=description,
            unit=unit,
            format=format,
            meta=meta,
            copy=copy,
            copy_indices=copy_indices,
        )

        if self.dtype == "object" and len(self) > 0:
            self._basedtype = type(self[0])
            assert all(isinstance(i, self.basedtype) for i in self)
        return self

    def __setitem__(self, index, value):
        array_value = np.array(value, ndmin=1)
        if self.basedtype and not all(
            isinstance(v, self.basedtype) for v in array_value
        ):
            raise ValueError(
                f"Value has not the same type {array_value.dtype} as this column {self.basedtype}"
            )
        return super().__setitem__(index, value)

    def _copy_attrs(self, obj):
        super()._copy_attrs(obj)
        if val := getattr(obj, "basedtype", None):
            self._basedtype = val

    @property
    def basedtype(self):
        return self._basedtype

    def astype(self, dtype, *args, **kwargs) -> "HColumn":
        res: "HColumn" = super().astype(dtype, *args, **kwargs)
        if np.dtype(dtype) != "O" and res.basedtype:
            res._basedtype = None

        elif np.dtype(dtype) == "O" and dtype is not self.basedtype:
            if isinstance(dtype, type) and issubclass(dtype, BasedReal):
                dtype = cast(Type[BasedReal], dtype)
                res._basedtype = dtype

                if self.basedtype:

                    def convert(x):
                        return dtype(x, self.significant)

                elif self.dtype == "int":
                    convert = dtype.from_int
                elif self.dtype == "float":

                    def convert(x):
                        return dtype.from_float(x, 3)

                else:
                    raise ValueError

                res.data[:] = np.vectorize(convert)(res.data)

        return res

    _base_repr_ = _patch_dtype_info_name(Column._base_repr_, 0)

    def truncate(self, significant: Optional[int] = None) -> "HColumn":
        return np.vectorize(lambda x: x.truncate(significant))(self)

    def ceil(self, significant: Optional[int] = None) -> "HColumn":
        return np.vectorize(lambda x: x.ceil(significant))(self)

    def floor(self, significant: Optional[int] = None) -> "HColumn":
        return np.vectorize(lambda x: x.floor(significant))(self)

    def __round__(self, significant: Optional[int] = None) -> "HColumn":
        return np.vectorize(lambda x: round(x, significant))(self)

    def resize(self, significant: int) -> "HColumn":
        return np.vectorize(lambda x: x.resize(significant))(self)

    @property
    def significant(self) -> Optional[int]:
        """
        If this column contains `BasedReal` values, return the minimum significant
        of all values. Else returns `None`
        """
        if self.basedtype:
            return min(x.significant for x in self)
        return None



from dataclasses import dataclass
from typing import List, Literal, Optional, Tuple

import pandas as pd
from pandas import DataFrame

from kanon.utils.types.number_types import Real

__all__ = ["Symmetry", "OutOfBoundsOriginError", "OverlappingSymmetryError"]


@dataclass
class Symmetry:
    """Defines a symmetry strategy that can be applied on a `~pandas.DataFrame`
    from a specified source interval to one or multiple target keys

    >>> df = DataFrame({"val": [5, 9, 2]} ,index=[0,1,3])
    >>> sym = Symmetry("mirror")
    >>> df.pipe(sym)
    val
    0   5
    1   9
    3   2
    5   9
    6   5
    >>> sym = Symmetry("periodic", sign=-1, offset=2)
    >>> df.pipe(sym)
    val
    0   5
    1   9
    3   2
    4  -3
    5  -7
    7   0
    >>> sym = Symmetry("periodic", sign=-1, source=(0,1), targets=[6,10])
    >>> df.pipe(sym)
    val
    0    5
    1    9
    3    2
    6   -5
    7   -9
    10  -5
    11  -9

    :param symtype: Type of the symmetry, it can be of the same direction (`periodic`) \
    or the oposite (`mirror`)
    :type symtype: Literal["periodic", "mirror"]
    :param offset: Offset to add to the symmetry values, defaults to 0
    :type offset: int, optional
    :param sign: Relative signs of the symmetry values from source values, defaults to 1
    :type sign: Literal[-1, 1], optional
    :param source: Tuple representing the lower and upper bound to take the values from, \
    defaults to the whole DataFrame
    :type source: Tuple[Real, Real], optional
    :param targets: List of keys where the symmetry are pasted, defaults to the end of \
    the DataFrame
    :type targets: List[int], optional
    """

    symtype: Literal["periodic", "mirror"]
    offset: int = 0
    sign: Literal[-1, 1] = 1
    source: Optional[Tuple[Real, Real]] = None
    targets: Optional[List[Real]] = None

    def __post_init__(self):
        if self.symtype not in ("periodic", "mirror"):
            raise ValueError
        if self.source:
            if self.source[0] >= self.source[1]:
                raise ValueError

    def __call__(self, df: DataFrame):

        if len(df) == 0:
            return df

        if self.source:
            if not (
                self.source[0] < df.index[-1] >= self.source[1]
                and self.source[0] >= df.index[0] < self.source[1]
            ):
                raise OutOfBoundsOriginError
            symdf = df.loc[self.source[0] : self.source[1]].copy()  # type: ignore
        else:
            symdf = df.copy()

        def apply(x):
            return self.sign * x + self.offset

        if not self.targets:

            if self.symtype == "mirror":
                symdf.index = symdf.index.map(lambda x: 2 * symdf.index[-1] - x)
                symdf = symdf[:-1][::-1]

            elif self.symtype == "periodic":
                symdf.index = symdf.index.map(
                    lambda x: 1 + x + symdf.index[-1] - symdf.index[0]
                )

            if self.sign == -1 or self.offset:
                symdf = symdf.applymap(apply)

            df = pd.concat([df, symdf])
        else:
            for t in self.targets:
                tdf = symdf.copy()

                if self.symtype == "mirror":
                    tdf.index = tdf.index.map(lambda x: tdf.index[-1] - x + t)
                    tdf = tdf[::-1]

                else:
                    tdf.index = tdf.index.map(lambda x: t + x - tdf.index[0])

                if self.sign == -1 or self.offset:
                    tdf = tdf.applymap(apply)

                if len(df.index.intersection(tdf.index)) > 0:
                    raise OverlappingSymmetryError

                df = pd.concat([df, tdf])

        return df.sort_index()


class OutOfBoundsOriginError(IndexError):
    """
    Catches applying on a DataFrame a symmetry with source values outside
    the DataFrame bounds
    """


class OverlappingSymmetryError(ValueError):
    """
    Catches applying on a DataFrame a symmetry with overlapping results
    """



import pytest

from kanon.models.meta import (
    TableType,
    _registered_ids,
    dmodel,
    get_model_by_id,
    models,
)


@pytest.fixture
def mock_models():

    _models = models.copy()
    __reg = _registered_ids.copy()

    models.clear()
    _registered_ids.clear()

    yield

    models.clear()
    _registered_ids.clear()
    models.update(_models)
    _registered_ids.update(__reg)


def test_table_type(mock_models):
    class A(TableType):
        C = "v"

    assert A.C in models


def test_dmodel(mock_models):
    class A(TableType):
        C = "c"
        R = "r"

    @dmodel(A.C, 5, 1)
    def func(x, a):
        ...

    assert get_model_by_id(5) == func
    assert models[A.C]["func"] == func

    assert func.formula_id == 5
    assert func.table_type == A.C
    assert func.params == (1,)
    assert func.args == 1

    with pytest.raises(ValueError) as err:

        @dmodel(A.C, 6, 1)
        def func1(x):
            ...

    assert "Incorrect number of parameters" in str(err)

    with pytest.raises(ValueError) as err:

        @dmodel(A.R, 5, 1)
        def func2(x, a):
            ...

    assert "Formula ID" in str(err)

    with pytest.raises(ValueError) as err:

        @dmodel(A.C, 8, 11, 5)
        def func(x, a, b):
            ...

    assert "func already" in str(err)


def test_astro_id():
    for tt in models:
        assert tt.astro_id() > 0



from enum import Enum, EnumMeta, unique
from inspect import signature
from typing import Dict, Protocol, Tuple


class ModelCallable(Protocol):
    def __call__(self, *args: float) -> float:
        ...

    __name__: str
    table_type: "TableType"
    args: int
    params: Tuple[int, ...]
    formula_id: int


models: Dict["TableType", Dict[str, ModelCallable]] = {}


class EnumRegisterMeta(EnumMeta):
    def __new__(*args, **kwargs):
        cls = EnumMeta.__new__(*args, **kwargs)
        models.update({v: {} for v in cls.__members__.values()})
        return cls


@unique
class TableType(Enum, metaclass=EnumRegisterMeta):
    @classmethod
    def astro_id(cls) -> int:
        """Returns its DISHAS Astronomical Object ID

        :rtype: int
        """
        raise NotImplementedError


_registered_ids: Dict[int, ModelCallable] = {}


def dmodel(table_type: TableType, formula_id: int, *params: int):
    """Decorator for models existing in DISHAS as a Formula Definition

    :param table_type: TableType of the formula
    :type table_type: TableType
    :param formula_id: ID of the formula from DISHAS
    :type formula_id: int
    :param *params: Parameters IDs
    :type formula_id: *int
    """

    def decorator(fn: ModelCallable):
        fn_params_len = len(signature(fn).parameters)

        if fn_params_len - len(params) not in (1, 2):
            raise ValueError("Incorrect number of parameters")

        if formula_id in _registered_ids:
            raise ValueError(f"Formula ID {formula_id} already registered")

        if fn.__name__ in models[table_type]:
            raise ValueError(
                f"{fn.__name__} already defined for table_type {table_type}"
            )

        models[table_type][fn.__name__] = fn
        _registered_ids[formula_id] = fn

        fn.table_type = table_type
        fn.args = fn_params_len - len(params)
        fn.params = params
        fn.formula_id = formula_id

        return fn

    return decorator


def get_model_by_id(formula_id: int) -> ModelCallable:
    """Returns a model function from its DISHAS Formula ID

    :type formula_id: int
    :rtype: ModelCallable
    """
    return _registered_ids[formula_id]



"""
In this module we define RadixBase and BasedReal.
RadixBase is the basis of the way we work with different radices.
BasedReal are a class of Real numbers with a 1-1 relation with a RadixBase.

.. testsetup::

    >>> import builtins
    >>> builtins.Sexagesimal = radix_registry["Sexagesimal"]
    >>> builtins.Historical = radix_registry["Historical"]

>>> from kanon.units import RadixBase, radix_registry
>>> example_base = RadixBase([20, 5, 18], [24, 60], "example_base", [" ","u ","sep "])
>>> number = radix_registry["ExampleBase"]((8, 12, 3, 1), (23, 31))
>>> number
08 12u 3sep 01 ; 23,31
>>> float(number)
15535.979861111111

"""

import math
from decimal import Decimal
from fractions import Fraction
from functools import cached_property, lru_cache
from numbers import Number
from numbers import Real as _Real
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Sequence,
    SupportsFloat,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

import numpy as np
from astropy.units.core import Unit, UnitBase, UnitTypeError
from astropy.units.quantity import Quantity
from astropy.units.quantity_helper.converters import UFUNC_HELPERS
from astropy.units.quantity_helper.helpers import _d

from kanon.utils.list_to_tuple import list_to_tuple
from kanon.utils.looping_list import LoopingList

from .precision import PreciseNumber, PrecisionMode, TruncatureMode, set_precision

__all__ = ["RadixBase", "BasedReal", "radix_registry"]


radix_registry: Dict[str, Type["BasedReal"]] = {}
"""Registry containing all instanciated BasedReal classes."""


class RadixBase:
    """
    A class representing a numeral system. A radix must be specified at each position,
    by specifying an integer list for the integer positions, and an integer list for the
    fractional positions.
    """

    def __init__(
        self,
        left: Sequence[int],
        right: Sequence[int],
        name: str,
        integer_separators: Optional[Sequence[str]] = None,
    ):
        """
        Definition of a numeral system. A radix must be specified for each integer position
        (left argument) and for each fractional position (right argument).
        Note that the integer position are counted from right to left (starting from the ';'
        symbol and going to the left).

        :param left: Radix list for the integer part
        :param right: Radix list for the fractional part
        :param name: Name of this numeral system
        :param integer_separators: List of string separators, used for displaying the integer part of the number
        """
        assert left and right
        assert all(isinstance(x, int) for x in left)
        assert all(isinstance(x, int) for x in right)

        self.left: LoopingList[int] = LoopingList(left)
        self.right: LoopingList[int] = LoopingList(right)
        self.name = name
        if integer_separators is not None:
            if len(integer_separators) != len(left):
                raise ValueError
            self.integer_separators = LoopingList[str](integer_separators)
        else:
            self.integer_separators = LoopingList[str](
                ["," if x != 10 else "" for x in left]
            )

        self.mixed = any(x != left[0] for x in tuple(left) + tuple(right))

        # Build a class inheriting from BasedReal, that will use this RadixBase as
        # its numeral system.
        type_name = "".join(map(str.capitalize, self.name.split("_")))
        if type_name in radix_registry:
            raise ValueError(f"Name {type_name} already exists in registry")

        new_type = type(type_name, (BasedReal,), {"base": self})
        radix_registry[type_name] = new_type

        # Store the newly created BasedReal class
        self.type: Type[BasedReal] = new_type

    @overload
    def __getitem__(self, key: int) -> int:
        ...

    @overload
    def __getitem__(self, key: slice) -> LoopingList[int]:
        ...

    def __getitem__(self, key):
        """
        Return the radix at the specified position. Position 0 represents the last integer
        position before the fractional part (i.e. the position just before the ';' in sexagesimal
        notation, or just before the '.' in decimal notation). Positive positions represent
        the fractional positions, negative positions represent the integer positions.

        :param key: Position. <= 0 for integer part (with 0 being the right-most integer position), > 0 for fractional part
        :return: Radix at the specified position
        """
        if isinstance(key, slice):
            if not key.start or not key.stop:
                raise ValueError("RadixBase slices must have a start and a stop value")
            array = [self[i] for i in range(key.start, key.stop)]
            return LoopingList[str](array[:: key.step])

        if key <= 0:
            return self.left[key - 1]
        else:
            return self.right[key - 1]

    @lru_cache
    def factor_at_pos(self, pos: int) -> int:
        """
        Returns an int factor corresponding to a digit at position pos.
        This factor is an integer, when dealing with fractional position you should invert
        the result to find relevant factor.

        >>> Sexagesimal.base.factor_at_pos(-2)
        3600
        >>> Sexagesimal.base.factor_at_pos(0)
        1

        :param pos: Position of the digit
        :type pos: int
        :return: Factor at pos
        :rtype: int
        """
        factor = 1
        for i in range(abs(pos)):
            factor *= self[i if pos > 0 else -i]
        return factor


def ndigit_for_radix(radix: int) -> int:
    """
    Compute how many digits are needed to represent a position of
    the specified radix.

    >>> ndigit_for_radix(10)
    1
    >>> ndigit_for_radix(60)
    2

    :param radix:
    :return:
    """
    return int(np.ceil(np.log10(radix)))


class BasedReal(PreciseNumber, _Real):
    """
    Abstract class allowing to represent a value in a specific `RadixBase`.
    Each time a new `RadixBase` object is recorded, a new class inheriting from BasedReal
    is created and recorded in `radix_registry`.
    The `RadixBase` to be used will be placed in the class attribute `base`

    Class attributes:
       - base :        A `RadixBase` object (will be attributed dynamically to the children inheriting this class)
    """

    base: RadixBase
    """`RadixBase` of this BasedReal"""
    __left: Tuple[int, ...]
    __right: Tuple[int, ...]
    __remainder: Decimal
    __sign: Literal[-1, 1]
    __slots__ = ("base", "__left", "__right", "__remainder", "__sign")

    def __check_range(self):
        """
        Checks that the given values are in the range of the base and are integers.
        """
        if self.sign not in (-1, 1):
            raise ValueError("Sign should be -1 or 1")
        if not (isinstance(self.remainder, Decimal) and 0 <= self.remainder < 1):
            if self.remainder == 1:  # pragma: no cover
                self += (self.one() * self.sign) >> self.significant
            else:
                raise ValueError(
                    f"Illegal remainder value ({self.remainder}), should be a Decimal between [0.,1.["
                )
        # if self.base.factor_at_pos(len(self.left) - 1) > 1e+15 and self.remainder:
        #     warnings.warn("""
        #                 Integer part of this number exceeds floating point precision.
        #                 Calculations made with this number as a float might return incorrect results.
        #                 """)
        for x in self[:]:
            if isinstance(x, float):
                raise IllegalFloatError(x)
            if not isinstance(x, int):
                raise TypeError(f"{x} not an int")
        for i, s in enumerate(self[:]):
            if s < 0.0 or s >= self.base[i - len(self.left) + 1]:
                raise IllegalBaseValueError(
                    self.base, self.base[i - len(self.left) + 1], s
                )

    def __simplify_integer_part(self) -> int:
        """
        Remove the useless trailing zeros in the integer part and return how many were removed
        """
        count = 0
        for i in self.left[:-1]:
            if i != 0:
                break
            count += 1
        if count > 0:
            self.__left = self.left[count:]

        return count != 0

    @list_to_tuple
    def __new__(cls, *args, remainder=Decimal(0.0), sign=1) -> "BasedReal":
        """Constructs a number with a given radix.

        Arguments:

        - `str`

        >>> Sexagesimal("-2,31;12,30")
        -02,31 ; 12,30

        - 2 `Sequence[int]` representing integral part and fractional part

        >>> Sexagesimal((2,31), (12,30), sign=-1)
        -02,31 ; 12,30
        >>> Sexagesimal([2,31], [12,30])
        02,31 ; 12,30

        - a `BasedReal` with a significant number of digits,

        >>> Sexagesimal(Sexagesimal("-2,31;12,30"), 1)
        -02,31 ; 12 |r0.5

        - multiple `int` representing an integral number in current `base`

        >>> Sexagesimal(21, 1, 3)
        21,01,03 ;

        :param remainder: When a computation requires more precision than the precision \
        of this number, we store a :class:`~decimal.Decimal` remainder to keep track of it, defaults to 0.0
        :type remainder: ~decimal.Decimal, optional
        :param sign: The sign of this number, defaults to 1
        :type sign: int, optional
        :raises ValueError: Unexpected or illegal arguments
        :rtype: BasedReal
        """
        if cls is BasedReal:
            raise TypeError("Can't instanciate abstract class BasedReal")
        self: BasedReal = super().__new__(cls)
        self.__left = ()
        self.__right = ()
        self.__remainder = remainder
        self.__sign = sign
        if np.all([isinstance(x, int) for x in args]):
            return cls.__new__(cls, args, (), remainder=remainder, sign=sign)
        if len(args) == 2:
            if isinstance(args[0], BasedReal):
                if isinstance(args[0], cls):
                    return args[0].resize(args[1])
                return cls.from_decimal(args[0].decimal, args[1])
            if isinstance(args[0], tuple) and isinstance(args[1], tuple):
                self.__left = args[0]
                self.__right = args[1]
            else:
                raise ValueError("Incorrect parameters at BasedReal creation")
        elif len(args) == 1:
            if isinstance(args[0], str):
                return cls._from_string(args[0])
            raise ValueError(
                "Please specify a number of significant positions"
                if isinstance(args[0], BasedReal)
                else "Incorrect parameters at BasedReal creation"
            )
        else:
            raise ValueError("Incorrect number of parameter at BasedReal creation")

        self.__check_range()

        if self.__simplify_integer_part() or not self.left:
            return cls.__new__(
                cls,
                self.left or (0,),
                self.right,
                remainder=self.remainder,
                sign=self.sign,
            )

        return self

    @property
    def left(self) -> Tuple[int, ...]:
        """
        Tuple of values at integer positions

        >>> Sexagesimal(1,2,3).left
        (1, 2, 3)

        :rtype: Tuple[int, ...]
        """
        return self.__left

    @property
    def right(self) -> Tuple[int, ...]:
        """
        Tuple of values at fractional positions

        >>> Sexagesimal((1,2,3), (4,5)).right
        (4, 5)

        :rtype: Tuple[int, ...]
        """
        return self.__right

    @property
    def remainder(self) -> Decimal:
        """
        When a computation requires more significant figures than the precision of this number,
        we store a :class:`~decimal.Decimal` remainder to keep track of it

        >>> Sexagesimal(1,2,3, remainder=Decimal("0.2")).remainder
        Decimal('0.2')

        :return: Remainder of this `BasedReal`
        :rtype: ~decimal.Decimal
        """
        return self.__remainder

    @property
    def sign(self) -> Literal[-1, 1]:
        """
        Sign of this `BasedReal`

        >>> Sexagesimal(1,2,3, sign=-1).sign
        -1

        :rtype: Literal[-1, 1]
        """
        return self.__sign

    @property
    def significant(self) -> int:
        """
        Precision of this `BasedReal` (equals to length of fractional part)

        >>> Sexagesimal((1,2,3), (4,5)).significant
        2

        :rtype: int
        """
        return len(self.right)

    @cached_property
    def decimal(self) -> Decimal:
        """
        This `BasedReal` converted as a `~decimal.Decimal`

        >>> Sexagesimal((1,2,3), (15,36)).decimal
        Decimal('3723.26')

        :rtype: Decimal
        """
        value = Decimal(abs(int(self)))
        factor = Decimal(1)
        for i in range(self.significant):
            factor *= self.base.right[i]
            value += self.right[i] / factor

        value += self.remainder / factor
        return value * self.sign

    def to_fraction(self) -> Fraction:
        """
        :return: this `BasedReal` as a :class:`~fractions.Fraction` object.
        """
        return Fraction(self.decimal)

    @classmethod
    def from_fraction(
        cls,
        fraction: Fraction,
        significant: Optional[int] = None,
    ) -> "BasedReal":
        """
        :param fraction: a `~fractions.Fraction` object
        :param significant: significant precision desired
        :return: a `BasedReal` object computed from a Fraction
        """

        if not isinstance(fraction, Fraction):
            raise TypeError(f"Argument {fraction} is not a Fraction")

        num, den = fraction.as_integer_ratio()
        res: BasedReal = cls.from_decimal(
            Decimal(num) / Decimal(den), significant or 100
        )

        return res if significant else res.minimize_precision()

    def __repr__(self) -> str:
        """
        Convert to string representation.
        Note that this representation is rounded (with respect to the remainder attribute) not truncated

        :return: String representation of this number
        """
        res = ""
        if self.sign < 0:
            res += "-"

        for i in range(len(self.left)):
            if i > 0:
                res += self.base.integer_separators[i - len(self.left)]
            num = str(self.left[i])
            digit = ndigit_for_radix(self.base.left[i - len(self.left)])
            res += "0" * (digit - len(num)) + num

        res += " ; "

        for i in range(len(self.right)):
            num = str(self.right[i])
            digit = ndigit_for_radix(self.base.right[i])
            res += "0" * (digit - len(num)) + num

            if i < len(self.right) - 1:
                res += ","

        if self.remainder:
            res += f" |r{self.remainder:3.1f}"

        return res

    __str__ = __repr__

    @classmethod
    def _from_string(cls, string: str) -> "BasedReal":
        """
        Parses and instantiate a `BasedReal` object from a string

        >>> Sexagesimal('1, 12; 4, 25')
        01,12 ; 04,25
        >>> Historical('2r 7s 29; 45, 2')
        2r 07s 29 ; 45,02
        >>> Sexagesimal('0 ; 4, 45')
        00 ; 04,45

        :param string: `str` representation of the number
        :return: a new instance of `BasedReal`
        """

        if not isinstance(string, str):
            raise TypeError(f"Argument {string} is not a str")

        string = string.strip().lower()
        if len(string) == 0:
            raise EmptyStringException("String is empty")
        if string[0] == "-":
            sign = -1
            string = string[1:]
        else:
            sign = 1
        left_right = string.split(";")
        if len(left_right) < 2:
            left = left_right[0]
            right = ""
        elif len(left_right) == 2:
            left, right = left_right
        else:
            raise TooManySeparators("Too many separators in string")

        left = left.strip()
        right = right.strip()

        left_numbers: List[int] = []
        right_numbers: List[int] = []

        if len(right) > 0:
            right_numbers = [int(i) for i in right.split(",")]

        if len(left) > 0:
            rleft = left[::-1]
            for i in range(len(left)):
                separator = cls.base.integer_separators[-i - 1].strip().lower()
                if separator != "":
                    split = rleft.split(separator, 1)
                    if len(split) == 1:
                        rem = split[0]
                        break
                    value, rem = split
                else:  # pragma: no cover
                    value = rleft[0]
                    rem = rleft[1:]
                left_numbers.insert(0, int(value[::-1]))
                rleft = rem.strip()
                if len(rleft) == 1:
                    break
            left_numbers.insert(0, int(rleft[::-1]))

        return cls(left_numbers, right_numbers, sign=sign)

    def resize(self, significant: int) -> "BasedReal":
        """
        Resizes and returns a new `BasedReal` object to the specified precision

        >>> n = Sexagesimal('02, 02; 07, 23, 55, 11, 51, 21, 36')
        >>> n
        02,02 ; 07,23,55,11,51,21,36
        >>> n.remainder
        Decimal('0')
        >>> n1 = n.resize(4)
        >>> n1.right
        (7, 23, 55, 11)
        >>> n1.remainder
        Decimal('0.8560000000000000000000000000')
        >>> n1.resize(7)
        02,02 ; 07,23,55,11,51,21,36

        :param significant: Number of desired significant positions
        :return: Resized `BasedReal`
        """
        if significant == self.significant:
            return self
        if significant > self.significant:
            rem = type(self).from_decimal(
                self.sign * self.remainder, significant - self.significant
            )
            return type(self)(
                self.left,
                self.right + rem.right,
                remainder=rem.remainder,
                sign=self.sign,
            )
        if significant >= 0:
            remainder = type(self)(
                (), self.right[significant:], remainder=self.remainder
            )

            return type(self)(
                self.left,
                self.right[:significant],
                remainder=remainder.decimal,
                sign=self.sign,
            )

        raise NotImplementedError

    def __trunc__(self):
        return int(float(self.truncate(0)))

    def truncate(self, significant: Optional[int] = None) -> "BasedReal":
        """
        Truncate this BasedReal object to the specified precision

        >>> n = Sexagesimal('02, 02; 07, 23, 55, 11, 51, 21, 36')
        >>> n
        02,02 ; 07,23,55,11,51,21,36
        >>> n = n.truncate(3); n
        02,02 ; 07,23,55
        >>> n = n.resize(7); n
        02,02 ; 07,23,55,00,00,00,00

        :param n: Desired significant positions
        :return: Truncated BasedReal
        """
        if significant is None:
            significant = self.significant
        if significant > self.significant:
            return self
        left = self.left if significant >= 0 else self.left[:-significant]
        right = self.right[:significant] if significant >= 0 else ()
        return type(self)(left, right, sign=self.sign)

    def floor(self, significant: Optional[int] = None) -> "BasedReal":
        resized = self.resize(significant) if significant else self
        if resized.remainder == 0 or self.sign == 1:
            return resized.truncate()

        return resized._set_remainder(Decimal(0.5)).__round__()

    def ceil(self, significant: Optional[int] = None) -> "BasedReal":
        resized = self.resize(significant) if significant else self
        if resized.remainder == 0 or self.sign == -1:
            return resized.truncate()

        return resized._set_remainder(Decimal(0.5)).__round__()

    def minimize_precision(self) -> "BasedReal":
        """
        Removes unnecessary zeros from fractional part of this BasedReal.

        :return: Minimized BasedReal
        """
        if self.remainder > 0 or self.significant == 0 or self.right[-1] > 0:
            return self

        count = 0
        for x in self.right[::-1]:
            if x != 0:
                break
            count += 1

        return self.truncate(self.significant - count)

    def __lshift__(self, other: int) -> "BasedReal":
        """self << other

        :param other: Amount to shift this BasedReal
        :type other: int
        :return: Shifted number
        :rtype: BasedReal
        """
        return self.shift(-other)

    def __rshift__(self, other: int) -> "BasedReal":
        """self >> other

        :param other: Amount to shift this BasedReal
        :type other: int
        :return: Shifted number
        :rtype: BasedReal
        """
        return self.shift(other)

    def shift(self, i: int) -> "BasedReal":
        """
        Shifts number to the left (-) or the right (+).
        Prefer using >> and << operators (right-shift and left-shift).

        >>> Sexagesimal(3).shift(-1)
        03,00 ;
        >>> Sexagesimal(3).shift(2)
        00 ; 00,03

        :param i: Amount to shift this BasedReal
        :return: Shifted number
        :rtype: BasedReal
        """

        if i == 0:
            return self

        if self.base.mixed:
            raise NotImplementedError

        offset = len(self.left) if i > 0 else len(self.left) - i
        br_rem = self.from_decimal(self.remainder, max(0, offset - len(self[:])))

        left_right = (0,) * i + self[:] + br_rem.right

        left = left_right[:offset]
        right = left_right[offset : -i if -i > offset else None]

        return type(self)(left, right, remainder=br_rem.remainder, sign=self.sign)

    @lru_cache
    def subunit_quantity(self, i: int) -> int:
        """Convert this sexagesimal to the integer value from the specified fractional point.

        >>> number = Sexagesimal("1,0;2,30")

        Amount of minutes in `number`

        >>> number.subunit_quantity(1)
        3602

        Amount of zodiacal signs in `number`

        >>> number.subunit_quantity(-1)
        1

        :param i: Rank of the subunit to compute from.
        :type i: int
        :return: Integer amount of the specified subunit.
        :rtype: int
        """

        res = 0
        factor = 1
        for idx, v in enumerate(self.resize(max(0, i + 1))[i::-1]):
            res += v * factor
            factor *= self.base[i - idx]
        return self.sign * res

    def __round__(self, significant: Optional[int] = None):
        """
        Round this BasedReal object to the specified precision.
        If no precision is specified, the rounding is performed with respect to the
        remainder attribute.

        >>> n = Sexagesimal('02, 02; 07, 23, 55, 11, 51, 21, 36')
        >>> n
        02,02 ; 07,23,55,11,51,21,36
        >>> round(n, 4)
        02,02 ; 07,23,55,12

        :param significant: Number of desired significant positions
        :return: self
        """
        if significant is None:
            significant = self.significant
        n = self.resize(significant)
        if n.remainder >= 0.5:
            with set_precision(
                pmode=PrecisionMode.MAX, tmode=TruncatureMode.NONE, recording=False
            ):
                values = [0] * significant + [1]
                n += type(self)(values[:1], values[1:], sign=self.sign)
        return n.truncate(significant)

    def __getitem__(self, key):
        """
        Allow to get a specific position value of this BasedReal object
        by specifying an index. The position 0 corresponds to the right-most integer position.
        Negative positions correspond to the other integer positions, positive
        positions correspond to the fractional positions.

        :param key: desired index
        :return: value at the specified position
        """
        if isinstance(key, slice):
            array = self.left + self.right
            start = key.start + len(self.left) - 1 if key.start is not None else None
            stop = key.stop + len(self.left) - 1 if key.stop is not None else None
            return array[start : stop : key.step]

        if isinstance(key, int):
            if -len(self.left) < key <= 0:
                return self.left[key - 1]
            if self.significant >= key > 0:
                return self.right[key - 1]
            raise IndexError

        raise TypeError

    @classmethod
    def from_float(
        cls, floa: float, significant: int, remainder_threshold: float = 0.999999
    ) -> "BasedReal":
        """
        Class method to produce a new BasedReal object from a floating number

        >>> Sexagesimal.from_float(1/3, 4)
        00 ; 20,00,00,00

        :param floa: floating value of the number
        :param significant: precision of the number
        :return: a new BasedReal object
        """

        if not isinstance(floa, (int, float)):
            raise TypeError(f"Argument {floa} is not a float")

        integer_part = cls.from_int(int(floa), significant=significant)
        value = abs(floa - int(integer_part))

        right = [0] * significant

        factor = 1.0
        if value != 0:
            for i in range(significant):
                factor = cls.base.right[i]
                value *= factor
                if value - int(value) > remainder_threshold and value + 1 < factor:
                    value = int(value) + 1
                elif value - int(value) < 1 - remainder_threshold and any(
                    x != 0 for x in right
                ):
                    value = int(value)
                position_value = int(value)
                value -= position_value
                right[i] = position_value

        return cls(
            integer_part.left,
            tuple(right),
            remainder=Decimal(value),
            sign=-1 if floa < 0 else 1,
        )

    @classmethod
    def from_decimal(cls, dec: Decimal, significant: int) -> "BasedReal":
        """
        Class method to produce a new BasedReal object from a Decimal number

        >>> Sexagesimal.from_decimal(Decimal('0.1'), 4)
        00 ; 06,00,00,00

        :param dec: floating value of the number
        :param significant: precision of the number
        :return: a new BasedReal object
        """

        if not isinstance(dec, Decimal):
            raise TypeError(f"Argument {dec} is not a Decimal")

        integer_part = cls.from_int(int(dec), significant=significant)
        value = abs(dec - int(integer_part))

        right = [0] * significant

        factor = Decimal(1)
        for i in range(significant):
            factor = cls.base.right[i]
            value *= factor
            position_value = int(value)
            value -= position_value
            right[i] = position_value

        return cls(
            integer_part.left, tuple(right), remainder=value, sign=-1 if dec < 0 else 1
        )

    @classmethod
    def zero(cls, significant=0) -> "BasedReal":
        """
        Class method to produce a zero number of the specified precision

        >>> Sexagesimal.zero(7)
        00 ; 00,00,00,00,00,00,00

        :param significant: desired precision
        :return: a zero number
        """
        return cls((0,), (0,) * significant)

    @classmethod
    def one(cls, significant=0) -> "BasedReal":
        """
        Class method to produce a unit number of the specified precision

        >>> Sexagesimal.one(5)
        01 ; 00,00,00,00,00

        :param significant: desired precision
        :return: a unit number
        """
        return cls((1,), (0,) * significant)

    @classmethod
    @overload
    def range(cls, stop: int) -> Generator["BasedReal", None, None]:
        ...

    @classmethod
    @overload
    def range(cls, start: int, stop: int, step=1) -> Generator["BasedReal", None, None]:
        ...

    @classmethod
    def range(cls, *args, **kwargs) -> Generator["BasedReal", None, None]:
        """
        Range generator, equivalent to `range` builtin but yields `BasedReal` numbers.

        :yield: `BasedReal` integers.
        """
        for i in range(*args, **kwargs):
            yield cls.from_int(i)

    @classmethod
    def from_int(cls, value: int, significant=0) -> "BasedReal":
        """
        Class method to produce a new BasedReal object from an integer number

        >>> Sexagesimal.from_int(12, 4)
        12 ; 00,00,00,00

        :param value: integer value of the number
        :param significant: precision of the number
        :return: a new BasedReal object
        """

        if not np.issubdtype(type(value), np.integer):
            raise TypeError(f"Argument {value} is not an int")

        base = cls.base
        sign = -1 if value < 0 else 1
        value *= sign

        pos = 0
        int_factor = 1

        while value >= int_factor:
            int_factor *= base.left[-1 - pos]
            pos += 1

        left = [0] * pos

        for i in range(pos):
            int_factor //= base.left[-pos + i]
            position_value = value // int_factor
            value -= position_value * int_factor
            left[i] = position_value

        return cls(left, (0,) * significant, sign=sign)

    def __float__(self) -> float:
        """
        Compute the float value of this BasedReal object

        >>> float(Sexagesimal('01;20,00'))
        1.3333333333333333
        >>> float(Sexagesimal('14;30,00'))
        14.5

        :return: float representation of this BasedReal object
        """
        value = float(abs(int(self)))
        factor = 1.0
        for i in range(self.significant):
            factor /= self.base.right[i]
            value += factor * self.right[i]

        value += factor * float(self.remainder)
        return float(value * self.sign)

    def __int__(self) -> int:
        """
        Compute the int value of this BasedReal object
        """
        value = 0
        factor = 1
        for i in range(len(self.left)):
            value += factor * self.left[-i - 1]
            factor *= self.base.left[-i - 1]

        return value * self.sign

    def _truediv(self, _other: PreciseNumber) -> "BasedReal":

        other = cast(BasedReal, _other)

        if self.base.mixed:
            return self.from_float(float(self) / float(other), self.significant)

        max_significant = max(self.significant, other.significant)

        if self == 0:
            return self.zero(significant=max_significant)
        elif other == 1:
            return self
        elif other == -1:
            return -self
        elif other == 0:
            raise ZeroDivisionError

        sign = self.sign * other.sign

        q_res = self.zero(max_significant)
        right = list(q_res.right)

        numerator = abs(self)
        denominator = abs(other)

        q, r = divmod(numerator, denominator)

        q_res += q

        for i in range(0, max_significant):
            numerator = r * self.base.right[i]
            q, r = divmod(numerator, denominator)
            if q == self.base.right[i]:  # pragma: no cover
                q_res += 1
                r = self.zero()
                break
            right[i] = int(q)

        return type(self)(
            q_res.left, right, remainder=r.decimal / denominator.decimal, sign=sign
        )

    def _add(self, _other: PreciseNumber) -> "BasedReal":

        other = cast(BasedReal, _other)

        if self.decimal == -other.decimal:
            return self.zero()

        maxright = max(self.significant, other.significant)
        maxleft = max(len(self.left), len(other.left))
        va = self.resize(maxright)
        vb = other.resize(maxright)

        sign = va.sign if abs(va) > abs(vb) else vb.sign
        if sign < 0:
            va = -va
            vb = -vb

        maxlen = max(len(va[:]), len(vb[:]))
        values = (
            [v.sign * x for x in v[::-1]] + [0] * (maxlen - len(v[:])) for v in (va, vb)
        )
        numbers: List[int] = [a + b for a, b in zip(*values)] + [0]

        remainder = va.remainder * va.sign + vb.remainder * vb.sign
        fn = remainder if remainder >= 0 else remainder - 1
        remainder -= int(fn)
        numbers[0] += int(fn)

        for i, r in enumerate(numbers):
            factor = self.base[maxright - i]
            if r < 0 or r >= factor:
                numbers[i] = r % factor
                numbers[i + 1] += 1 if r > 0 else -1

        numbers = [abs(x) for x in numbers[::-1]]
        left = numbers[: maxleft + 1]
        right = numbers[maxleft + 1 :]

        return type(self)(left, right, remainder=abs(remainder), sign=sign)

    def __add__(self, other) -> "BasedReal":
        """
        self + other

        >>> Sexagesimal('01, 21; 47, 25') + Sexagesimal('45; 32, 14, 22')
        02,07 ; 19,39,22
        """
        if not np.isreal(other):
            raise NotImplementedError

        if type(self) is not type(other):
            return self + self.from_float(float(other), significant=self.significant)

        return super().__add__(other)

    def __radd__(self, other) -> "BasedReal":
        """other + self"""
        return self + other

    def _sub(self, _other: PreciseNumber) -> "BasedReal":

        other = cast(BasedReal, _other)
        return self + -other

    def __rtruediv__(self, other):
        """other / self"""
        return other / float(self)

    def __pow__(self, exponent):
        """self**exponent
        Negative numbers cannot be raised to a non-integer power
        """
        res = self.one(self.significant)

        if exponent == 0:
            return res
        if self == 0:
            return self

        if self < 0 and int(exponent) != exponent:
            raise ValueError(
                "Negative BasedReal cannot be raised to a non-integer power"
            )

        int_exp = int(exponent)
        f_exp = float(exponent - int_exp)

        if int_exp > 0:
            for _ in range(0, int_exp):
                res *= self
        else:
            for _ in range(0, -int_exp):
                res /= self
        res *= float(self) ** f_exp

        return res

    def __rpow__(self, base):
        """base ** self"""
        return self.from_float(float(base), self.significant) ** self

    def __neg__(self) -> "BasedReal":
        """-self"""
        return type(self)(
            self.left, self.right, remainder=self.remainder, sign=-self.sign
        )

    def __pos__(self) -> "BasedReal":
        """+self"""
        return self

    def __abs__(self) -> "BasedReal":
        """
        abs(self)

        >>> abs(Sexagesimal('-12; 14, 15'))
        12 ; 14,15

        :return: the absolute value of self
        """
        if self.sign >= 0:
            return self
        return -self

    def _mul(self, _other: PreciseNumber) -> "BasedReal":

        other = cast(BasedReal, _other)

        if self in (1, -1):
            return other if self == 1 else -other
        if other in (1, -1):
            return self if other == 1 else -self
        if self == 0 or other == 0:
            return self.zero()

        if self.base.mixed:
            return self.from_float(float(self) * float(other), self.significant)

        max_right = max(self.significant, other.significant)

        va = self.resize(max_right)
        vb = other.resize(max_right)

        res_int = int(va << max_right) * int(vb << max_right)

        res = self.from_int(res_int) >> 2 * max_right

        factor = self.base.factor_at_pos(max_right)
        vb_rem = vb.sign * vb.remainder / factor
        va_rem = va.sign * va.remainder / factor

        rem = (
            va.truncate().decimal * vb_rem
            + vb.truncate().decimal * va_rem
            + va_rem * vb_rem
        )

        if rem:
            res += float(rem)

        return res

    @overload
    def __mul__(self, other: Union[float, "BasedReal"]) -> "BasedReal":  # type: ignore
        ...

    @overload
    def __mul__(self, other: Unit) -> "BasedQuantity":
        ...

    def __mul__(self, other):
        """
        self * other

        >>> Sexagesimal('01, 12; 04, 17') * Sexagesimal('7; 45, 55')
        09,19 ; 39,15 |r0.7
        """

        if isinstance(other, UnitBase):
            return BasedQuantity(self, unit=other)

        if not np.isreal(other) or not isinstance(other, SupportsFloat):
            raise NotImplementedError

        if type(self) is not type(other):
            return self * self.from_float(float(other), self.significant)

        return super().__mul__(other)

    def __rmul__(self, other):
        """other * self"""
        return self * other

    def __divmod__(self, other: Any) -> Tuple["BasedReal", "BasedReal"]:
        """divmod(self, other)"""

        if type(self) is type(other):

            if self.base.mixed:
                res = divmod(float(self), float(other))
                return (
                    self.from_float(res[0], self.significant),
                    self.from_float(res[1], self.significant),
                )

            max_sig = max(self.significant, other.significant)
            if self == 0:
                zero = self.zero(max_sig)
                return (zero, zero)

            max_significant = max(self.significant, other.significant)
            s_self = self.resize(max_significant)
            s_other = other.resize(max_significant)
            if s_self.remainder == s_other.remainder == 0:
                qself = s_self.subunit_quantity(max_significant)
                qother = s_other.subunit_quantity(max_significant)
                fdiv, mod = divmod(qself, qother)
                return (
                    self.from_int(fdiv, max_sig),
                    self.from_int(mod) >> max_significant,
                )

            fdiv = math.floor(self.decimal / other.decimal)
            if fdiv == self.decimal / other.decimal:
                mod = Decimal(0)
            else:
                mod = self.decimal % other.decimal + (
                    0 if self.sign == other.sign else other.decimal
                )
            return self.from_int(fdiv, max_sig), self.from_decimal(mod, max_sig)

        if np.isreal(other):
            return divmod(self, self.from_float(float(other), self.significant))

        raise NotImplementedError

    def __floordiv__(self, other) -> "BasedReal":  # type: ignore
        """self // other"""
        return divmod(self, other)[0]

    def __rfloordiv__(self, other):
        """other // self: The floor() of other/self."""
        return other // float(self)

    def __mod__(self, other) -> "BasedReal":
        """self % other"""
        return divmod(self, other)[1]

    def __rmod__(self, other):
        """other % self"""
        return other % float(self)

    @overload
    def __truediv__(self, other: Number) -> "BasedReal":  # type: ignore
        ...

    @overload
    def __truediv__(self, other: Unit) -> "BasedQuantity":
        ...

    def __truediv__(self, other) -> "BasedReal":
        """self / other"""
        if isinstance(other, UnitBase):
            return self * (other ** -1)

        if type(self) is type(other):
            return super().__truediv__(other)

        return self / self.from_float(float(other), significant=self.significant)

    def __gt__(self, other) -> bool:
        """self > other"""
        if not isinstance(other, Number):
            return other <= self
        if isinstance(other, BasedReal):
            return self.decimal > other.decimal
        other = cast(SupportsFloat, other)
        return float(self) > float(other)

    def __eq__(self, other) -> bool:
        """self == other"""
        if not isinstance(other, SupportsFloat):
            return False
        if isinstance(other, BasedReal):
            return self.decimal == other.decimal
        return float(self) == float(other)

    def equals(self, other: "BasedReal") -> bool:
        """Tests strict equivalence between this BasedReal and another

        >>> Sexagesimal("1,2;3").equals(Sexagesimal("1,2;3"))
        True
        >>> Sexagesimal("1,2;3").equals(Sexagesimal("1,2;3,0"))
        False

        :param other: The other BasedReal to be compared with the first
        :type other: BasedReal
        :return: True if both objects are the same, False otherwise
        :rtype: bool

        """
        if type(self) is not type(other):
            return False

        return (
            self.left == other.left
            and self.right == other.right
            and self.sign == other.sign
            and self.remainder == other.remainder
        )

    def __ne__(self, other) -> bool:
        """self != other"""
        return not self == other

    def __ge__(self, other) -> bool:
        """self >= other"""
        return self > other or self == other

    def __lt__(self, other) -> bool:
        """self < other"""
        return not self >= other

    def __le__(self, other) -> bool:
        """self <= other"""
        return not self > other

    def __floor__(self):
        """Finds the greatest Integral <= self."""
        return self.__trunc__() + (1 if self.sign < 0 else 0)

    def __ceil__(self):
        """Finds the least Integral >= self."""
        return self.__trunc__() + (1 if self.sign > 0 else 0)

    def __hash__(self) -> int:
        if self.remainder == 0 and all([x == 0 for x in self.right]):
            return int(self)
        return hash((self.left, self.right, self.sign, self.remainder))

    def sqrt(self, iteration: Optional[int] = None) -> "BasedReal":
        """Returns the square root, using Babylonian method

        :param iteration: Number of iterations, defaults to the significant number
        :type iteration: Optional[int], optional
        """
        if self.sign < 0:
            raise ValueError("Square root domain error")

        if self == 0:
            return self

        if iteration is None:
            iteration = self._get_significant(self)

        if self >= 1:
            res = self.from_int(int(math.sqrt(float(self))))
        else:
            res = self.from_float(math.sqrt(float(self)), self.significant)
            iteration = 0

        for _ in range(iteration):
            res += self / res
            res /= 2

        return res

    def _set_remainder(self, remainder: Decimal) -> "BasedReal":
        return type(self)(self.left, self.right, sign=self.sign, remainder=remainder)


class BasedQuantity(Quantity):

    value: BasedReal

    def __new__(cls, value, unit, **kwargs):
        if (
            not isinstance(value, BasedReal)
            or isinstance(value, (Sequence, np.ndarray))
            and not all(isinstance(v, BasedReal) for v in value)
        ):
            return Quantity(value, unit, **kwargs)

        def _len(_):
            del type(value).__len__
            return 0

        type(value).__len__ = _len
        self = super().__new__(cls, value, unit=unit, dtype=object, **kwargs)
        return self

    def __mul__(self, other) -> "BasedQuantity":  # pragma: no cover
        return super().__mul__(other)

    def __add__(self, other) -> "BasedQuantity":  # pragma: no cover
        return super().__add__(other)

    def __sub__(self, other) -> "BasedQuantity":  # pragma: no cover
        return super().__sub__(other)

    def __truediv__(self, other) -> "BasedQuantity":  # pragma: no cover
        return super().__truediv__(other)

    def __lshift__(self, other) -> "BasedQuantity":
        if isinstance(other, Number):
            return super(Quantity, self).__lshift__(other)
        return super().__lshift__(other)

    def __rshift__(self, other) -> "BasedQuantity":
        if isinstance(other, Number):
            return super(Quantity, self).__rshift__(other)
        return super().__rshift__(other)

    def __getattr__(self, attr: str):
        if attr.startswith(("_", "__")) and not attr.endswith("__"):
            raise AttributeError
        vect = np.frompyfunc(lambda x: getattr(x, attr), 1, 1)
        properties = ("left", "right", "significant", "sign", "remainder", "base")
        unit = _d(self.unit) if attr not in properties else None
        UFUNC_HELPERS[vect] = lambda *_: ([None, None], unit)
        if callable(getattr(BasedReal, attr)):

            def _new_func(*args):
                vfunc = np.frompyfunc(lambda x: x(*args), 1, 1)
                UFUNC_HELPERS[vfunc] = lambda *_: ([None, None], unit)
                return vfunc(vect(self))

            return _new_func
        return vect(self)

    def __round__(self, significant: Optional[int] = None) -> "BasedQuantity":
        return self.__getattr__("__round__")(significant)

    def __abs__(self) -> "BasedQuantity":  # pragma: no cover
        return self.__getattr__("__abs__")()

    def __quantity_subclass__(self, _):
        return type(self), True


# here we define standard bases and automatically generate the corresponding BasedReal classes
RadixBase([60], [60], "sexagesimal")
RadixBase([10, 12, 30], [60], "historical", ["", "r ", "s "])
RadixBase([10], [100], "historical_decimal")
RadixBase([10], [60], "integer_and_sexagesimal")
RadixBase([10], [24, 60], "temporal")
# add new definitions here, corresponding BasedReal inherited classes will be automatically generated


def _shift_helper(f, unit1, unit2):
    if unit2:  # pragma: no cover
        raise UnitTypeError(
            "Can only apply '{}' function to "
            "dimensionless quantities".format(f.__name__)
        )
    return [None, None], _d(unit1)


UFUNC_HELPERS[np.left_shift] = _shift_helper
UFUNC_HELPERS[np.right_shift] = _shift_helper


class BasedRealException(Exception):
    pass


class EmptyStringException(BasedRealException, ValueError):
    pass


class TooManySeparators(BasedRealException, ValueError):
    pass


class IllegalBaseValueError(BasedRealException, ValueError):
    """
    Raised when a value is not in the range of the specified base.

    ```python
    if not 0 <= val < radix[i]:
        raise IllegalBaseValueError(radix, radix[i], val)
    ```
    """

    def __init__(self, radix, base, num):
        super().__init__()
        self.radix = radix
        self.base = base
        self.num = num

    def __str__(self):
        return f"An invalid value for ({self.radix.name}) was found \
        ('{self.num}'); should be in the range [0,{self.base}[)."


class IllegalFloatError(BasedRealException, TypeError):
    """
    Raised when an expected int value is a float.

    ```python
    if isinstance(val, float):
        raise IllegalFloatError(val)
    ```
    """

    def __init__(self, num):
        super().__init__()
        self.num = num

    def __str__(self):
        return f"An illegal float value was found ('{self.num}')"



"""
Common interpolation methods are defined in this module.

There are 2 types of interpolation functions :

Single-Point Interpolators, which interpolate on a single value
    `linear_interpolation`

    `quadratic_interpolation`

Whole Interpolators, which interpolate on every `NaN` value
    `distributed_interpolation`
"""

from functools import wraps
from typing import Callable, Literal, Tuple

import numpy as np
import pandas as pd
from scipy.interpolate import lagrange

from kanon.utils.types.number_types import Real

__all__ = [
    "Interpolator",
    "linear_interpolation",
    "quadratic_interpolation",
    "distributed_interpolation",
]


Interpolator = Callable[[pd.DataFrame, Real], Real]


def _split_df(df: pd.DataFrame, key: Real) -> Tuple[pd.DataFrame, pd.DataFrame]:

    df = df.rename_axis("x")
    df = df.rename(columns={list(df.columns)[0]: "y"})

    df = df.reset_index().set_index("x", drop=False)

    lower = df.truncate(after=key)
    upper = df.truncate(before=key)

    return lower, upper


def _interpolation_decorator(func):
    """
    This decorator automatically casts the key in the correct type and returns the result
    if the key is in the DataFrame
    """

    @wraps(func)
    def wrapper(df: pd.DataFrame, key: Real) -> Real:
        if df.index.dtype == "object" and isinstance(key, float):
            key = type(df.index[0]).from_float(key, df.index[0].significant)
        if key in df.index:
            return df.loc[key][0]
        return func(df, key)

    return wrapper


@_interpolation_decorator
def linear_interpolation(df: pd.DataFrame, key: Real) -> Real:
    """Linear interpolation.
    Will prioritize taking the lower and upper value.
    The `pd.DataFrame` needs at least 2 rows.
    """

    assert len(df) >= 2, "The DataFrame needs at least 2 rows"

    lower, upper = _split_df(df, key)

    if len(lower) == 0:
        (_, a), (_, b) = upper.iloc[:2].T.iteritems()
    elif len(upper) == 0:
        (_, a), (_, b) = lower.iloc[-2:].T.iteritems()
    else:
        a = lower.iloc[-1]
        b = upper.iloc[0]

    c = (b.y - a.y) / (b.x - a.x)

    return c * (key - a.x) + a.y


@_interpolation_decorator
def quadratic_interpolation(df: pd.DataFrame, key: Real) -> Real:
    """Quadratic interpolation, from Lagrange
    Will prioritize taking 2 values before the keys and 1 after.
    The `pd.DataFrame` needs at least 3 rows.
    """

    assert len(df) >= 3, "The DataFrame needs at least 3 rows"

    lower, upper = _split_df(df, key)

    before = max(min(3 - min(len(upper), 1), len(lower)), 3 - len(upper))
    after = 3 - before

    values = pd.concat([lower.iloc[-before:], upper.iloc[:after]])

    poly = lagrange(list(values["x"]), list(values["y"]))

    return poly(key)


# Whole DataFrame interpolation
# Interpolates on every NaN value


def distributed_interpolation(
    df: pd.DataFrame, direction: Literal["convex", "concave"]
):
    """Applies distributed interpolation on a `DataFrame` with a regularly stepped index.
    Interpolates on every unknown values (`numpy.nan` or `pandas.NA`).
    """

    df = df.copy()

    if direction not in ("convex", "concave"):
        raise ValueError(
            f"The interpolation direction must be either convex or concave, not {direction}"
        )

    if pd.isna(df.iloc[-1][0]) or pd.isna(df.iloc[0][0]):
        raise ValueError("The DataFrame must start and end with non nan values")

    if based_values := df.iloc[0].dtypes == "object":

        based_type = type(df.iloc[0][0])

        based_idx = df[~df.isna().any(axis=1)].index

        max_sig: int = df.loc[based_idx].applymap(lambda x: x.significant).max().iloc[0]
        df.loc[based_idx] = df.loc[based_idx].applymap(
            lambda x: x.subunit_quantity(max_sig)
        )

        df = df.astype(float)

    if df.isna().sum()[0] < len(df) - 2:

        def edges(x: pd.Series) -> float:
            if np.isnan(x).sum() == 1:
                return 1
            return np.nan

        bounds = df.rolling(2, 1).apply(edges).dropna().index

        for b in range(0, len(bounds), 2):
            lower = df.index.get_loc(bounds[b]) - 1
            upper = df.index.get_loc(bounds[b + 1]) + 1
            df.iloc[lower:upper] = distributed_interpolation(
                df.iloc[lower:upper], direction=direction
            )

    else:

        index_diff = df.index.to_series().diff().iloc[1:].to_numpy()
        step = index_diff[0]

        if not (index_diff == step).all():
            raise ValueError("The DataFrame must have regular steps")

        first: Real = df.iloc[0][0]
        last: Real = df.iloc[-1][0]

        q, r = divmod(last - first, len(df) - 1)

        r = r if direction == "concave" else r - len(df) + 2

        for idx, _ in df.iloc[1:-1].iterrows():
            first += q + (1 if r > 0 else 0)

            r += 1 if direction == "convex" else -1

            df.loc[idx] = first

    if based_values:
        df.loc[:] = df.applymap(lambda x: based_type.from_int(int(x)).shift(max_sig))

    return df



from functools import partial
from typing import (
    Callable,
    Generic,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import numpy as np
import pandas as pd
from astropy.table import Row, Table
from astropy.table.operations import join
from astropy.table.pprint import TableFormatter
from astropy.table.table import TableAttribute
from astropy.units import Quantity
from astropy.units.core import Unit

from kanon.tables.hcolumn import HColumn, _patch_dtype_info_name
from kanon.utils.types.number_types import Real

from .interpolations import (
    Interpolator,
    distributed_interpolation,
    linear_interpolation,
)
from .symmetries import Symmetry

__all__ = ["HTable"]


T = TypeVar("T")


class GenericTableAttribute(TableAttribute, Generic[T]):
    def __get__(self, instance, owner) -> T:
        return super().__get__(instance, owner)


class HTableFormatter(TableFormatter):

    _pformat_col = _patch_dtype_info_name(TableFormatter._pformat_col, 1)


class HTable(Table):
    """`HTable` is a subclass of `astropy.table.Table`, made to model Historical Astronomy tables
    representing mathematical functions. Its argument column or columns are its index, while the
    values should be on the first column. Columns are allowed to contain all kinds of `~numbers.Real`,
    especially `~kanon.units.radices.BasedReal` numbers. `HTable` also provides additional historical
    features and metadata.

    See also: https://docs.astropy.org/en/stable/table/

    >>> table = HTable({"args": [1,2,3], "values": [5.1,3.9,4.3]}, index="args")
    >>> table
    <HTable length=3>
    args  values
    int64 float64
    ----- -------
        1     5.1
        2     3.9
        3     4.3
    >>> table.loc[2]
    <Row index=1>
    args  values
    int64 float64
    ----- -------
        2     3.9
    >>> table.loc[2]["values"]
    3.9

    :param data: Data to initialize table.
    :type data: Optional[Data]
    :param names: Specify column names.
    :type names: Union[List[str], Tuple[str, ...]]
    :param index: Columns considered as the indices.
    :type index: Optional[Union[str, List[str]]]
    :param units: List or dict of units to apply to columns.
    :type units: Optional[List[Unit]]
    :param dtype: Specify column data types.
    :type dtype: Optional[List]
    :param symmetry: Specify a list of `~kanon.tables.Symmetry` on this table. Defaults to empty list.
    :type symmetry: Optional[List[Symmetry]]
    :param interpolate: Specify a custom interpolation method,\
    defaults to `~kanon.tables.interpolations.linear_interpolation`.
    :type interpolate: Optional[Interpolator]
    :param opposite: Defines if the table values should be of the opposite sign. Defaults to False.
    :type opposite: Optional[bool]

    """

    Column = HColumn
    TableFormatter = HTableFormatter

    interpolate = GenericTableAttribute[Interpolator](default=linear_interpolation)
    """Interpolation method."""
    symmetry: List[Symmetry] = TableAttribute(default=[])
    """Table symmetries."""
    opposite: bool = TableAttribute(default=False)
    """Defines if the table values should be of the opposite sign."""

    def __init__(
        self,
        data=None,
        names: Optional[Union[List[str], Tuple[str, ...]]] = None,
        index: Optional[Union[str, List[str]]] = None,
        dtype: Optional[List] = None,
        units: Optional[List[Unit]] = None,
        *args,
        **kwargs,
    ):

        super().__init__(
            data=data, names=names, units=units, dtype=dtype, *args, **kwargs
        )

        if index:
            self.set_index(index)

    def _check_index(self, index=None):
        if not self.indices and not index:
            raise IndexError(
                "HTable should have an index, defining the function's arguments"
            )
        return self.primary_key[0]

    def to_pandas(
        self, index=None, use_nullable_int=True, symmetry=True
    ) -> pd.DataFrame:
        self._check_index(index)
        df = super().to_pandas(index=index, use_nullable_int=use_nullable_int)
        if symmetry:
            for sym in self.symmetry:
                df = df.pipe(sym)
        return df

    @property
    def is_double(self) -> bool:
        """Is this table a double argument table"""
        return self.columns[self.values_column].dtype.name == "void128"

    @overload
    def get(self, key: Union[Real, Quantity], with_unit: Literal[False]) -> Real:
        ...

    @overload
    def get(self, key: Real, with_unit: Literal[True] = True) -> Union[Real, Quantity]:
        ...

    @overload
    def get(self, key: Quantity, with_unit: Literal[True] = True) -> Quantity:
        ...

    def get(self, key: Union[Real, Quantity], with_unit=True):
        """Get the value from any key based on interpolated data.

        :param key: Argument for an interpolated function
        :type key: `~numbers.Real`
        :param with_unit: Whether the result is represented as a Quantity or not. \
        Defaults to `True`
        :type with_unit: bool
        :raises IndexError: Key is out of bounds
        :return: Interpolated value
        :rtype: `~numbers.Real`
        """

        if self.is_double:
            return HTable(
                self.loc[key][self.values_column],
                index=self.meta["index"],
                units=[self.meta["unit_arg"], self.meta["unit_entry"]]
                if "unit_arg" in self.meta
                else None,
                meta=self.meta.copy(),
            )

        df = self.to_pandas()

        unit = (self.columns[self.values_column].unit if with_unit else 1) or 1

        val: Real

        if isinstance(key, int):
            val = float(key)

        elif isinstance(key, Quantity):
            val = key.to(self[self.primary_key[0]].unit).value

        else:
            val = key

        return self.interpolate(df, val) * unit

    def apply(
        self, column: str, func: Callable, new_name: Optional[str] = None
    ) -> "HTable":
        """
        Applies a function on a column and returns the new `HTable`

        :param column: Name of the column to be modified
        :param func: Function applied to the column
        :param new_name: New name for the modified column, optional
        :return: HTable with modified column
        :rtype: HTable
        """

        table = self.copy()
        try:
            table[column] = func(table[column])
        except (TypeError, AttributeError, ValueError):
            table[column] = np.vectorize(func)(table[column])

        if new_name:
            table.rename_columns([column], [new_name])

        return table

    def set_index(self, index: Union[str, List[str]], engine=None):
        for c in self.colnames:
            self.remove_indices(c)

        self.add_index(index, unique=True, engine=engine)

    @property
    def values_column(self) -> str:
        """
        Column representing the values
        """
        self._check_index()

        set_val = set(self.colnames) - set(self.primary_key)
        assert len(set_val) == 1, f"Values should be in 1 column, not {len(set_val)}"

        return list(set_val)[0]

    def copy(self, set_index=None, copy_data=True) -> "HTable":
        table: HTable = super().copy(copy_data=copy_data)
        if set_index:
            table.set_index(set_index)
        return table

    def populate(
        self, array: List[Real], method: Literal["mask", "interpolate"] = "mask"
    ) -> "HTable":
        """
        Populate a table with values from new keys contained in `array`.
        """

        key = self._check_index()

        right = {key: array}

        if method == "interpolate":
            right[self.values_column] = [self.get(x) for x in array]

        right_tab = HTable(right)
        if (index_type := self[key].basedtype) != right_tab[key].basedtype:
            right_tab[key] = right_tab[key].astype(index_type or self[key].dtype)
            if index_type is not None:
                right_tab[key] = right_tab[key].resize(self[key].significant)

        table: HTable = join(self, right_tab, join_type="outer")

        table.set_index(key)

        return table

    def fill(
        self,
        method: Union[
            Literal["distributed_convex", "distributed_concave"],
            Callable[[pd.DataFrame], pd.DataFrame],
        ],
        bounds: Optional[Tuple[Real, Real]] = None,
    ) -> "HTable":
        """Fill masked values within `bounds` with the specified `method`.
        To fill with a unique value, see `~HTable.filled`.

        :param method: Method to use for filling masked values
        :type method: Literal["distributed_convex", "distributed_concave"]
        :param bounds: Tuple of arguments bounds, defaults to the whole table
        :type bounds: Optional[Tuple[Real, Real]], optional
        """

        slice_bounds = slice(*(bounds or (None, None)))

        filltab: HTable = self.loc[slice_bounds]

        if isinstance(filltab, Row):
            return self.copy()

        valcol = filltab[filltab.values_column]

        if valcol[0] is np.ma.masked or valcol[-1] is np.ma.masked:
            raise ValueError("First and last rows must not be masked")

        if np.ma.masked not in valcol:
            return self.copy()

        fill_method: Callable[[pd.DataFrame], pd.DataFrame]

        if method == "distributed_concave":
            fill_method = partial(distributed_interpolation, direction="concave")
        elif method == "distributed_convex":
            fill_method = partial(distributed_interpolation, direction="convex")
        elif callable(method):
            fill_method = method
        else:
            raise ValueError("Incorrect fill method")

        df = filltab.to_pandas().pipe(fill_method)

        tab_copy = self.copy()
        tab_copy.loc[slice_bounds] = HTable.from_pandas(df, index=True)

        if not np.ma.is_masked(tab_copy[tab_copy.values_column]):
            tab_copy[tab_copy.values_column] = HColumn(tab_copy[filltab.values_column])

        return tab_copy

    def diff(
        self, n=1, prepend: List = [], append: List = [], new_name: Optional[str] = None
    ) -> "HTable":
        """Applies `np.diff` on this table's column values to calculate the n-th difference.
        By default, it will prepend the first value.

        :return: HTable with n-th difference values.
        :rtype: HTable
        """

        cat = prepend + append
        if len(cat) > n or n > len(cat) > 0:
            raise ValueError(
                f"You must prepend or append exactly n ({n}) values or 0 values"
            )

        if len(cat) == 0:
            prepend = [self[0][self.values_column]] * n

        kwargs = {}
        if prepend:
            kwargs["prepend"] = prepend
        if append:
            kwargs["append"] = append

        diff = partial(np.diff, n=n, **kwargs)

        return self.apply(self.values_column, diff, new_name)

    def plot2d(self, *args, **kwargs):
        """
        Plot this `HTable` with the argument column as X and values column as Y.
        Uses the same arguments as `matplotlib.pyplot.plot`.
        """

        assert (
            len(self.primary_key) == 1
        ), "plot2d is only valid with single argument tables"

        from matplotlib import pyplot as plt

        x: HColumn = self[self.primary_key[0]]
        y: HColumn = self[self.values_column]

        plt.xlabel(f"{x.name} ({x.unit})" if x.unit else x.name)
        plt.ylabel(f"{y.name} ({y.unit})" if y.unit else y.name)

        return plt.plot(x, y, *args, **kwargs)


def join_multiple(
    *tables,
    keys=None,
    join_type="inner",
    uniq_col_name="{col_name}_{table_name}",
    table_names=["1", "2"],
    metadata_conflicts="silent",
    join_funcs=None,
) -> Table:
    """
    `astropy.table.join` but with more than one table, iteratively from left to right.
    """

    _join = partial(
        join,
        keys=keys,
        join_type=join_type,
        uniq_col_name=uniq_col_name,
        table_names=table_names,
        metadata_conflicts=metadata_conflicts,
        join_funcs=join_funcs,
    )

    new_table = _join(*tables[:2])
    for t in tables[2:]:
        new_table = _join(new_table, t)

    return new_table



import hypothesis.strategies as st
import numpy as np
import pandas as pd
import pytest
from hypothesis.core import given

from kanon.tables.interpolations import (
    distributed_interpolation,
    linear_interpolation,
    quadratic_interpolation,
)


def build_dataframe():
    return pd.DataFrame([5, 4, 6, 12], [1, 5, 7, 8])


class TestInterpolations:
    @given(st.floats(allow_infinity=False, allow_nan=False, min_value=1, max_value=8))
    def test_linear_hypothesis(self, key: float):
        df = build_dataframe()

        np_interp = np.interp(key, np.array(df.index), np.array(df[0]))

        assert linear_interpolation(df, key) == np_interp

    def test_linear(self):
        df = build_dataframe()

        assert linear_interpolation(df, -1) == 5.5
        assert linear_interpolation(df, 9) == 18

    def test_quadratic(self):
        df = build_dataframe()

        assert quadratic_interpolation(df, 1.25) == 4.7421875
        assert np.isclose(quadratic_interpolation(df, 10), 34)

    def test_distributed(self):
        df = pd.DataFrame([5] + [np.nan] * 8 + [45], list(range(6, 16)))

        convex = df.pipe(distributed_interpolation, direction="convex")
        assert list(convex[0]) == [5, 9, 13, 17, 21, 25, 30, 35, 40, 45]
        concave = df.pipe(distributed_interpolation, direction="concave")
        assert list(concave[0]) == [5, 10, 15, 20, 25, 29, 33, 37, 41, 45]

        df = pd.DataFrame([5] + [np.nan] * 6 + [45], [i / 4 for i in range(8)])
        concave = df.pipe(distributed_interpolation, direction="concave")
        assert len(concave) == 8 and not np.isnan(concave[0]).any()

        df = pd.DataFrame(
            [5] + [np.nan] * 6 + [45] + [np.nan] * 6 + [100], [i / 4 for i in range(15)]
        )
        concave = df.pipe(distributed_interpolation, direction="concave")
        assert len(concave) == 15 and not np.isnan(concave[0]).any()

        with pytest.raises(ValueError) as err:
            df = pd.DataFrame([5] + [np.nan] * 8 + [45], list(range(6, 15)) + [31])
            df.pipe(distributed_interpolation, direction="convex")
        assert str(err.value) == "The DataFrame must have regular steps"

        with pytest.raises(ValueError) as err:
            df = pd.DataFrame([5] + [np.nan] * 9, list(range(6, 16)))
            df.pipe(distributed_interpolation, direction="convex")
        assert str(err.value) == "The DataFrame must start and end with non nan values"

        with pytest.raises(ValueError) as err:
            df.pipe(distributed_interpolation, direction="unknown")
        assert "unknown" in str(err.value)

        from kanon.units import Sexagesimal

        df = pd.DataFrame(
            [Sexagesimal(5)] + [np.nan] * 8 + [Sexagesimal(45)], list(range(6, 16))
        )

        convex = df.pipe(distributed_interpolation, direction="convex")
        assert list(convex[0]) == [5, 9, 13, 17, 21, 25, 30, 35, 40, 45]

        df = pd.DataFrame(
            [Sexagesimal("0;0,5")] + [np.nan] * 8 + [Sexagesimal("0;0,45")],
            list(range(6, 16)),
        )

        convex = df.pipe(distributed_interpolation, direction="convex")
        assert list(convex[0]) == [
            Sexagesimal.from_int(x).shift(2)
            for x in [5, 9, 13, 17, 21, 25, 30, 35, 40, 45]
        ]
        assert convex.index[0].dtype == "int64"



from typing import Generic, TypeVar

T = TypeVar("T")


class LoopingList(list, Generic[T]):
    """
    A class for lists looping both sides.
    If an index is queried outside the boundaries of the list, the last element is returned.

    >>> test = LoopingList([0, 1, 2])
    >>> test
    [..., 0, 1, 2, ...]
    >>> test[45]
    2
    >>> test[-5]
    0

    """

    def __getitem__(self, key):
        if key >= len(self):
            return self[-1]
        if key < -len(self) + 1:
            return self[0]
        return super().__getitem__(key)

    def __repr__(self) -> str:
        return f"[..., {super().__repr__()[1:-1]}, ...]" if self else "[]"



from typing import Any, Dict, List, Literal, Optional, TypedDict, Union

# flake8: noqa


class OriginalValue(TypedDict):
    value: List[str]
    comment: str
    suggested: bool
    critical_apparatus: List[str]


NumberType = Literal[
    "sexagesimal", "floating sexagesimal", "integer and sexagesimal", "historical"
]
SymmetryType = Literal["mirror", "periodic"]
SymmetryOperation = Literal["opposite", "identity", "addition", "substraction"]
UnitType = Literal["degree", "day", "hour", "month", "year"]


class TemplateArgs(TypedDict):
    name: str
    type: NumberType
    unit: int
    nsteps: int
    ncells: int
    decpos: int
    subUnit: Optional[NumberType]
    firstMonth: Optional[Any]


class TemplateEntries(TypedDict):
    name: str
    type: NumberType
    ncells: int
    decpos: int
    unit: int


class Template(TypedDict):

    table_type: str
    readonly: bool
    args: List[TemplateArgs]
    entries: List[TemplateEntries]


class DSymmetry(TypedDict):
    symtype: SymmetryType
    offset: int
    sign: Literal[-1, 1]
    source: Optional[List[List[str]]]
    target: Optional[List[List[str]]]


class Args(TypedDict):
    argument1: List[str]
    argument2: List[str]


class ValueFloat(TypedDict):
    args: Args
    entry: List[str]
    template: Template


class OriginalArgs(TypedDict):
    argument1: List[OriginalValue]
    argument2: List[OriginalValue]


class SourceValueOriginal(TypedDict):
    args: OriginalArgs
    entry: Any


class ValueOriginal(SourceValueOriginal):
    template: Template
    symmetries: List


class TableContent(TypedDict):
    argument1_number_of_steps: int
    argument2_number_of_steps: int
    id: int
    source_value_original: SourceValueOriginal
    value_original: ValueOriginal
    value_float: ValueFloat
    corrected_value_float: ValueFloat
    edited_text: Dict
    entry_type_of_number: NumberType
    entry_number_unit: UnitType
    entry_significant_fractional_place: str
    entry_number_of_cell: str
    argument1_name: str
    argument1_type_of_number: NumberType
    argument1_number_unit: UnitType
    argument1_number_of_cell: str
    argument1_significant_fractional_place: str
    argument2_name: str
    argument2_type_of_number: NumberType
    argument2_number_unit: UnitType
    argument2_number_of_cell: str
    argument2_significant_fractional_place: str
    comment: str
    mathematical_parameter: str
    parameter_sets: Dict
    table_type: Dict
    created: str
    updated: str
    public: Literal[True]
    mean_motion: Optional[Any]
    kibana_name: str
    kibana_id: str
    symmetries: List[DSymmetry]



import math as m
import operator as op
import warnings
from decimal import Decimal, InvalidOperation
from fractions import Fraction

import hypothesis
import pytest
from hypothesis import strategies as st
from hypothesis.core import given

from kanon.units import BasedReal, Historical, Sexagesimal
from kanon.units.radices import (
    EmptyStringException,
    IllegalBaseValueError,
    IllegalFloatError,
    TooManySeparators,
)


def test_init():
    assert (
        Sexagesimal((1, 2, 31), (6,), sign=-1, remainder=Decimal("0.3")).__repr__()
        == "-01,02,31 ; 06 |r0.3"
    )

    # From float
    assert Sexagesimal.from_float(-0.016666666666666666, 2) == -Sexagesimal((0,), (1,))
    assert Sexagesimal.from_float(0.5, 4).equals(Sexagesimal("0; 30, 0, 0, 0"))
    with pytest.raises(TypeError):
        Sexagesimal.from_float("s", 1)

    # From int
    assert Sexagesimal.from_int(5, 2) == Sexagesimal(5)
    with pytest.raises(TypeError):
        Sexagesimal.from_int("s")

    # From Decimal
    assert Sexagesimal.from_decimal(Decimal(5), 2) == Sexagesimal(5)
    with pytest.raises(TypeError):
        Sexagesimal.from_decimal(5, 1)

    # From Fraction
    assert Sexagesimal.from_fraction(Fraction(5, 1)).equals(Sexagesimal(5))
    assert Sexagesimal.from_fraction(Fraction(5, 2)) == Sexagesimal("2;30")
    with pytest.raises(TypeError):
        Sexagesimal.from_fraction(5)

    # From str
    assert Sexagesimal("21,1,6,3;34") == Sexagesimal((21, 1, 6, 3), (34,))
    assert Sexagesimal("0,0,0,0;").equals(Sexagesimal(0))
    with pytest.raises(TypeError):
        Sexagesimal._from_string(5)
    with pytest.raises(EmptyStringException):
        Sexagesimal("")
    with pytest.raises(TooManySeparators):
        Sexagesimal("1;2;3")

    # From Sequence
    with pytest.raises(IllegalBaseValueError) as err:
        Sexagesimal((-6, 3), ())
    assert "should be in the range" in str(err.value)
    with pytest.raises(IllegalFloatError) as err:
        Sexagesimal((0.3, 5), (6, 8))
    assert "An illegal float" in str(err.value)

    # From BasedReal

    assert Sexagesimal(Historical("3;15"), 1).equals(Sexagesimal("3;15"))

    # From multiple ints
    with pytest.raises(ValueError):
        Sexagesimal(3, 5, remainder=Decimal(-5))
    with pytest.raises(ValueError):
        Sexagesimal(3, 5, remainder=0.6)

    # Incorrect parameters
    with pytest.raises(TypeError):
        BasedReal()
    with pytest.raises(ValueError):
        Sexagesimal(Decimal(5))
    with pytest.raises(ValueError):
        Sexagesimal("", "")
    with pytest.raises(ValueError):
        Sexagesimal("", "", "", "")
    with pytest.raises(TypeError):
        Sexagesimal(("a", 2), (1, 2))
    with pytest.raises(ValueError):
        Sexagesimal(1, sign=2)


def test_get():
    s = Sexagesimal("1, 2, 30; 18, 12, 23")
    assert s[-2] == 1
    assert s[0] == 30
    assert s[1] == 18
    assert s[:] == (1, 2, 30, 18, 12, 23)
    assert s[:0] == (1, 2)
    assert s[3::-1] == (23, 12, 18, 30, 2, 1)
    assert s[-1:] == (2, 30, 18, 12, 23)
    assert s[-1:2] == (2, 30, 18)

    with pytest.raises(IndexError):
        s[100]
    with pytest.raises(TypeError):
        s["5"]


def test_truncations():
    s = Sexagesimal("1, 2, 30; 18, 52, 23")
    assert round(s).equals(s)
    assert round(s, 1).equals(Sexagesimal("1,2,30;19"))
    assert s.truncate(1).equals(Sexagesimal("1,2,30;18"))
    assert s.truncate(100).equals(s)
    assert m.trunc(s) == 3750
    assert m.floor(s) == 3750
    assert m.ceil(s) == 3751
    assert m.floor(-s) == -3749
    assert m.ceil(-s) == -3750
    assert Sexagesimal(1, 2, 3).minimize_precision().equals(Sexagesimal(1, 2, 3))
    assert (
        Sexagesimal("1, 2, 3; 0, 0").minimize_precision().equals(Sexagesimal(1, 2, 3))
    )


def test_misc():
    s = Sexagesimal(5)
    assert (s or 1) == 5
    assert not s.equals("5")
    assert s.to_fraction() == Fraction(5)

    assert 5 / s == 1

    assert s ** -1 == 1 / 5
    assert s ** 1 == s
    assert 1 ** s == 1
    assert Sexagesimal(0) ** 1 == 0

    with pytest.raises(ValueError):
        (-Sexagesimal(5)) ** 2.5

    assert s > 4

    assert (s / 1).equals(s)
    assert (s / -1).equals(-s)
    with pytest.raises(ZeroDivisionError):
        s / 0

    assert Sexagesimal("1,0;2,30,1").subunit_quantity(1) == 3602

    assert 5 % Sexagesimal(2) == 1
    assert 5 // Sexagesimal(2) == 2

    assert divmod(Sexagesimal(5), 3) == divmod(5, 3)


def test_shift():
    s = Sexagesimal("20, 1, 2, 30; 0")
    assert (s >> 1).equals(Sexagesimal("20, 1, 2; 30, 0"))
    assert (s << 1).equals(Sexagesimal("20, 1, 2, 30, 0"))
    assert (s >> -1).equals(s << 1)
    assert (s >> 7).equals(Sexagesimal("0; 0, 0, 0, 20, 1, 2, 30, 0"))
    s = Sexagesimal((20,), (0, 2, 0), remainder=Decimal(0.5))
    assert (s << 2).equals(Sexagesimal((20, 0, 2), (0,), remainder=Decimal(0.5)))
    assert (s << 5).equals(Sexagesimal(20, 0, 2, 0, 30, 0))


@given(st.integers(min_value=0, max_value=15), st.integers(min_value=0, max_value=15))
def test_resize(x, y):
    s = Sexagesimal("1, 2, 3; 30, 25")
    assert s.resize(x) == s
    resized = s.resize(x).resize(y)
    assert m.isclose(resized, s)
    assert resized.significant == y


@given(st.floats(allow_infinity=False, allow_nan=False))
def test_comparisons(x):
    s = Sexagesimal("1, 2; 30")
    xs = Sexagesimal.from_float(x, 1)
    for comp in (op.lt, op.le, op.eq, op.ne, op.ge, op.gt):
        if comp(float(s), x):
            assert comp(s, xs)
        else:
            assert not comp(s, xs)


def biop_testing(x: BasedReal, y: BasedReal, operator):
    fx, fy = float(x), float(y)
    if (
        operator == op.pow
        and (fy < 0 or fy > 10 or x < 0 and int(fy) != fy)
        or operator == op.truediv
        and fy == 0
    ):
        return
    a = float(operator(x, y))
    b: float = operator(fx, fy)
    try:
        abstol = 1e-09 if a and b else 1e-11
        assert m.isclose(a, b.real, abs_tol=abstol)
        a = float(
            operator(x, type(x).from_float(fy, x.significant, remainder_threshold=1))
        )
        b = float(
            operator(type(x).from_float(fx, y.significant, remainder_threshold=1), y)
        )
        assert m.isclose(a, b.real, abs_tol=abstol)
    except Exception as e:
        hypothesis.note(f"{x.remainder} {y.remainder}")
        hypothesis.note(f"{operator.__name__}: {a} {b}")
        raise e


@given(st.from_type(Sexagesimal), st.from_type(Sexagesimal))
def test_operations_with_remainders(x, y):
    fx = float(x)

    assert float(-x) == -fx
    assert float(+x) == fx
    assert abs(x) == abs(fx)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for o in (op.mul, op.add, op.sub):
            biop_testing(x, y, o)

    if y != 0:
        try:
            biop_testing(x, y, op.truediv)
        except InvalidOperation:
            pass


@given(st.from_type(Sexagesimal), st.from_type(Sexagesimal))
def test_operations_without_remainders(x, y):
    x, y = x.truncate(), y.truncate()

    for o in (op.mul, op.add, op.sub, op.pow, op.truediv):
        biop_testing(x, y, o)


@given(
    st.integers(min_value=int(-1e15), max_value=int(1e15)).map(Sexagesimal.from_int),
    st.integers(min_value=int(-1e15), max_value=int(1e15))
    .filter(lambda x: x != 0)
    .map(Sexagesimal.from_int),
)
def test_mod_integers(x, y):
    hypothesis.assume(int(x) % int(y) == float(x) % float(y))
    biop_testing(x, y, op.mod)
    biop_testing(x, y, op.floordiv)


def test_sqrt():

    assert Sexagesimal(9).sqrt().equals(Sexagesimal(3))

    assert Sexagesimal(0).sqrt() == 0

    assert Sexagesimal("12;15").sqrt(5) == 3.5

    with pytest.raises(ValueError):
        Sexagesimal("-5").sqrt()


@given(st.from_type(Sexagesimal).filter(lambda x: x > 0))
def test_sqrt_hypo(n):
    assert m.isclose(float(n.sqrt(5)), m.sqrt(float(n)))


@given(
    st.integers(-20, 20),
    st.integers(-20, 20),
    st.integers(-2, 2).filter(lambda x: x != 0),
)
def test_range(start: int, stop: int, step: int):
    based_range = Sexagesimal.range(start, stop, step)
    normal_range = range(start, stop, step)

    assert list(normal_range) == [int(x) for x in based_range]

    based_range_stop = Sexagesimal.range(stop)
    normal_range_stop = range(stop)

    assert list(normal_range_stop) == [int(x) for x in based_range_stop]


def test_mixed_misc():
    assert Historical.from_int(60) == Historical("2s0")
    assert divmod(Historical(5), Historical("1s26;3")) == (0, 5)
    assert round(Historical(5, remainder=Decimal(0.6))) == 6


@given(st.from_type(Historical), st.from_type(Historical))
def test_operations_without_remainders_mixed(x, y):
    x, y = x.truncate(), y.truncate()

    for o in (op.mul, op.add, op.sub, op.pow, op.truediv):
        biop_testing(x, y, o)


def test_mod():
    assert (Sexagesimal("1,2;3,4") % Sexagesimal(4)).significant == 2
    val = Sexagesimal((30,), (30,), remainder=Decimal("0.9"))
    assert -val % 360 == 360 - val



from typing import Callable, Dict, List, Tuple, Union, cast

import astropy.units as u
import requests
from astropy.units.core import Unit

from kanon.tables.symmetries import Symmetry
from kanon.units import Historical, Sexagesimal
from kanon.units.radices import BasedReal
from kanon.utils.types.dishas import DSymmetry, NumberType, TableContent, UnitType
from kanon.utils.types.number_types import Real

from .htable import HTable

_dishas_fields = '","'.join(
    [
        "source_value_original",
        "argument1_name",
        "argument1_number_unit",
        "argument1_significant_fractional_place",
        "argument1_type_of_number",
        "argument2_name",
        "argument2_number_unit",
        "argument2_significant_fractional_place",
        "argument2_type_of_number",
        "entry_significant_fractional_place",
        "entry_number_unit",
        "entry_type_of_number",
        "symmetries",
        "edited_text",
    ]
)
DISHAS_REQUEST_URL = f'https://dishas.obspm.fr/elastic-query\
?index=table_content&hits=true&id={{}}&source=["{_dishas_fields}"]'


def read_sexag_array(array: List[str], shift: int) -> BasedReal:
    sign = -1 if array[0][0] == "-" else 1
    return Sexagesimal(*(abs(int(v)) for v in array), sign=sign) >> shift


def read_intsexag_array(array: List[str], _: int) -> BasedReal:
    integer = int(array[0])
    if len(array) == 1:
        return Sexagesimal.from_int(integer)
    return integer + (read_sexag_array(array[1:], 0) >> len(array) - 1)


def read_historical(array: List[str], shift: int) -> BasedReal:
    integer = int(array[0])

    # Special case for non true Historical in DISHAS with 2 values
    if len(array) == 2 and shift == 0:
        return integer * 30 + Historical(array[1])

    values = tuple(int(x) for x in array)
    return Historical(
        values[: -shift or None],
        values[-shift or len(values) :],
        sign=1 if integer >= 0 else -1,
    )


number_reader: Dict[NumberType, Callable[[List[str], int], Real]] = {
    "sexagesimal": read_sexag_array,
    "floating sexagesimal": read_sexag_array,
    "integer and sexagesimal": read_intsexag_array,
    "historical": read_historical,
}

unit_reader: Dict[UnitType, Unit] = {
    "degree": u.degree,
    "day": u.day,
    "year": u.year,
    "month": u.year / 12,
    "hour": u.hour,
}


def read_table_dishas(
    requested_id: Union[int, str], symmetry=True, units=True, entries_name="Entries"
) -> HTable:

    qid = int(requested_id)

    res: TableContent = requests.get(
        DISHAS_REQUEST_URL.format(qid),
    ).json()
    if not res or "error" in res:
        raise FileNotFoundError(f"{qid} ID not found in DISHAS database")

    return read_table_content(res, symmetry, units, entries_name)


def read_table_content(
    tabc: TableContent, symmetry=True, units=True, entries_name="Entries"
):

    values = tabc["source_value_original"]

    arg_unit = tabc["argument1_number_unit"]
    arg_shift = int(tabc["argument1_significant_fractional_place"])
    arg_reader = number_reader.get(tabc["argument1_type_of_number"], lambda x, _: x)

    entry_unit = tabc["entry_number_unit"]
    entry_shift = int(tabc["entry_significant_fractional_place"])

    def entry_reader(val):
        reader = number_reader.get(tabc["entry_type_of_number"], lambda x, _: x)
        if "**" in val:
            return reader(["0"], entry_shift)
        return reader(val, entry_shift)

    args = [arg_reader(v["value"], arg_shift) for v in values["args"]["argument1"]]

    # Double argument

    if "argument2" in values["args"]:
        len1 = len(values["args"]["argument1"])
        len2 = len(values["args"]["argument2"])

        arg2_unit = tabc["argument2_number_unit"]
        arg2_shift = int(tabc["argument2_significant_fractional_place"])
        arg2_reader = number_reader.get(
            tabc["argument2_type_of_number"], lambda x, _: x
        )

        args2 = [
            arg2_reader(v["value"], arg2_shift) for v in values["args"]["argument2"]
        ]

        tables = [
            HTable(
                [
                    args,
                    [
                        entry_reader(v["value"])
                        for v in [values["entry"][j][i] for j in range(len1)]
                    ],
                ],
                names=(tabc["argument1_name"], entries_name),
                index=(tabc["argument1_name"]),
                dtype=[object, object],
            )
            for i in range(len2)
        ]

        units_info = (
            {
                "unit_arg": unit_reader.get(arg_unit),
                "unit_entry": unit_reader.get(entry_unit),
            }
            if units
            else {}
        )

        return HTable(
            [args2, tables],
            names=(tabc["argument2_name"], "Tables"),
            index=(tabc["argument2_name"]),
            units=[unit_reader.get(arg2_unit), u.dimensionless_unscaled]
            if units
            else None,
            dtype=None,
            meta={
                **tabc["edited_text"],
                **units_info,
                "index": tabc["argument1_name"],
                "double": True,
            },
        )

    entries = [entry_reader(v["value"]) for v in values["entry"]]

    symmetry_raw = tabc["symmetries"]

    def build_sym(data: DSymmetry) -> Symmetry:
        symmetry = Symmetry(data["symtype"], data["offset"], sign=data["sign"])

        if source_data := data["source"]:
            source = tuple(arg_reader(v, arg_shift) for v in source_data)
            assert len(source) == 2
            symmetry.source = cast(Tuple[Real, Real], source)

        if target_data := data["target"]:
            symmetry.targets = [arg_reader(v, arg_shift) for v in target_data]

        return symmetry

    symmetries = [build_sym(sym) for sym in symmetry_raw] if symmetry else []

    table = HTable(
        [args, entries],
        names=(tabc["argument1_name"], entries_name),
        index=(tabc["argument1_name"]),
        units=[unit_reader.get(arg_unit), unit_reader.get(entry_unit)]
        if units
        else None,
        dtype=[object, object],
        symmetry=symmetries,
        meta=tabc["edited_text"],
    )

    return table



import json

import pytest
import requests_mock as reqmock
from astropy.utils.data import get_pkg_data_filename


@pytest.fixture(autouse=True)
def mock_request_dishas(requests_mock: reqmock.Mocker):
    error = {"error": "Non existing id"}

    def callback(req, _):
        try:
            qid = req.query.split("&id=")[1].split("&source")[0]
            path = get_pkg_data_filename(f"data/table_content-{qid}.json")
            with open(path, "r") as f:
                content = json.load(f)
        except (OSError, IndexError):
            content = error
        return content

    requests_mock.get(reqmock.ANY, json=callback)



"""The `precision` module is used when wanting to adjust `~kanon.units.radices.BasedReal`
arithmetical operations behavior.
All operations are made within a `PrecisionContext` rules, which indicate :

- A `TruncatureMode`
- A `PrecisionMode`
- 4 `CustomArithmeticAlgorithm`, (add, sub, mul, div)

Default precision context is set to `TruncatureMode.NONE`, `PrecisionMode.MAX`, and all
`CustomArithmeticAlgorithm` as default.

To set new precision rules you should use the `set_precision` context manager. In the example
below, I set the precision so that the result significant number is 0 and that it should be
truncated.

>>> from kanon.units import Sexagesimal
>>> a = Sexagesimal("1;50")
>>> b = Sexagesimal("2;0")
>>> a + b
03 ; 50
>>> with set_precision(tmode=TruncatureMode.TRUNC, pmode=0):
...     a + b
...
03 ;

If you want to use a specific algorithm for one of the arithmetical operations,
you first need to define the algorithm with this signature with the `identify_func`
decorator specifying a unique ID :

.. code:: python

    Callable[[PreciseNumber, PreciseNumber], PreciseNumber]

For example, a multiplication algorithm which is essentialy equivalent to ``a * round(b,0)``
with `TEST_MUL` as its ID:

>>> @identify_func("TEST_MUL")
... def test_mul(a: PreciseNumber, b: PreciseNumber) -> PreciseNumber:
...     res = 0
...     for i in range(int(round(b,0))):
...         res += a
...     return res

You can now use this function to make multiplications use this algorithm

>>> b * a
03 ; 40
>>> with set_precision(mul=test_mul):
...     b * a
04 ; 00

All operations and their associated context are stored inside the `ContextPrecision` when
the recording flag is set to ``True``. You can either set it to ``True`` inside of a
`set_precision` context manager, or globally turn it on with `set_recording(True)`.
Records are displayed with `get_records`, and can be cleared with `clear_records`.

Let's try to record our operations.

>>> set_recording(True)
>>> get_records()
[]
>>> a + b
03 ; 50
>>> with set_precision(tmode=TruncatureMode.ROUND, pmode=1):
...     a + Sexagesimal("2;5,30")
03 ; 56
>>> with set_precision(mul=test_mul):
...     b * a
04 ; 00
>>> get_records()
[{'args': (01 ; 50, 02 ; 00, '+', 03 ; 50), 'tmode': 'NONE', 'pmode': 'MAX', 'add': \
'DEFAULT', 'sub': 'DEFAULT', 'mul': 'DEFAULT', 'div': 'DEFAULT'}, {'args': (01 ; 50, \
02 ; 05,30, '+', 03 ; 56), 'tmode': 'ROUND', 'pmode': 1, 'add': 'DEFAULT', 'sub': \
'DEFAULT', 'mul': 'DEFAULT', 'div': 'DEFAULT'}, {'args': (02 ; 00, 01 ; 50, '*', 04 ; \
00), 'tmode': 'NONE', 'pmode': 'MAX', 'add': 'DEFAULT', 'sub': 'DEFAULT', 'mul': \
'TEST_MUL', 'div': 'DEFAULT'}]
>>> clear_records()
>>> set_recording(False)
>>> a + b
03 ; 50
>>> get_records()
[]

Types
-----

.. py:attribute:: CustomArithmeticAlgorithm

    :type: Optional[Callable[[PreciseNumber, PreciseNumber], PreciseNumber]
"""
from __future__ import annotations

import abc
from contextlib import contextmanager
from dataclasses import asdict, dataclass, field
from enum import Enum
from functools import partial, wraps
from numbers import Number
from typing import Callable, Dict, List, Literal, Optional, Union

__all__ = [
    "PrecisionMode",
    "TruncatureMode",
    "set_precision",
    "PrecisionContext",
    "PreciseNumber",
    "get_context",
    "set_context",
    "set_recording",
    "get_records",
    "clear_records",
    "CustomArithmeticAlgorithm",
    "Truncable",
]


def _with_context_precision(func=None, symbol=None):
    if not func:
        return partial(_with_context_precision, symbol=symbol)

    @wraps(func)
    def wrapper(*args, **kwargs) -> "Truncable":

        with set_precision(recording=False):
            value: "Truncable" = func(*args, **kwargs)

        if len(args) != 2 or any(not isinstance(a, Truncable) for a in args):
            return value

        if not isinstance(value, Truncable):
            raise TypeError

        value = value.resize(args[0]._get_significant(args[1]))
        ctx = get_context()
        value = ctx.tmode(value)
        ctx.record(*args, symbol, value)
        return value

    return wrapper


class Truncable(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def resize(self, significant: int) -> "Truncable":
        raise NotImplementedError

    @abc.abstractmethod
    def truncate(self, significant: Optional[int] = None) -> "Truncable":
        raise NotImplementedError

    @abc.abstractmethod
    def ceil(self, significant: Optional[int] = None) -> "Truncable":
        raise NotImplementedError

    @abc.abstractmethod
    def floor(self, significant: Optional[int] = None) -> "Truncable":
        raise NotImplementedError

    @abc.abstractmethod
    def __round__(self, significant: Optional[int] = None) -> "Truncable":
        raise NotImplementedError


class PreciseNumber(Number, Truncable):
    """Abstract class of numbers with `PrecisionContext` compatibility"""

    @property
    @abc.abstractmethod
    def significant(self) -> int:
        raise NotImplementedError

    def _get_significant(self, other: "PreciseNumber") -> int:
        return get_context()._precisionfunc(self, other)

    @_with_context_precision(symbol="+")
    def __add__(self, other):
        if f := get_context().add:
            return f(self, other)
        return self._add(other)

    @abc.abstractmethod
    def _add(self, other: "PreciseNumber") -> "PreciseNumber":
        raise NotImplementedError

    @_with_context_precision(symbol="-")
    def __sub__(self, other):
        if f := get_context().sub:
            return f(self, other)
        return self._sub(other)

    @abc.abstractmethod
    def _sub(self, other: "PreciseNumber") -> "PreciseNumber":
        raise NotImplementedError

    @_with_context_precision(symbol="*")
    def __mul__(self, other):
        if f := get_context().mul:
            return f(self, other)
        return self._mul(other)

    @abc.abstractmethod
    def _mul(self, other: "PreciseNumber") -> "PreciseNumber":
        raise NotImplementedError

    @_with_context_precision(symbol="/")
    def __truediv__(self, other):
        if f := get_context().div:
            return f(self, other)
        return self._truediv(other)

    @abc.abstractmethod
    def _truediv(self, other: "PreciseNumber") -> "PreciseNumber":
        raise NotImplementedError

    @abc.abstractmethod
    def __float__(self) -> float:
        raise NotImplementedError


class FuncEnum(Enum):
    def __call__(self, *args, **kwds) -> PreciseNumber:
        return self.value[0](*args, **kwds)


class PrecisionMode(FuncEnum):
    """Enumeration of standard precision modes available.
    You can also use a positive integer to indicate a precision at a constant significant number.
    """

    SCI = (
        lambda x, y: min(x.significant, y.significant),
        0,
    )  #: Following scientific notation
    MAX = (lambda x, y: max(x.significant, y.significant), 1)  #: Using max significant
    FULL = (
        lambda *_: (_ for _ in ()).throw(NotImplementedError),
        2,
    )  #: TODO Full calculation


class TruncatureMode(FuncEnum):
    """Enumeration of standard truncature modes available."""

    NONE = (lambda x: x, 0)  #: No truncature
    ROUND = (round, 1)  #: round()
    TRUNC = (lambda x: x.truncate(), 2)  #: truncate()
    CEIL = (lambda x: x.ceil(), 3)  #: ceil()
    FLOOR = (lambda x: x.floor(), 4)  #: floor()


CustomArithmeticAlgorithm = Optional[
    Callable[[PreciseNumber, PreciseNumber], PreciseNumber]
]

_AI_REGISTRY: Dict[CustomArithmeticAlgorithm, str] = {None: "DEFAULT"}


def identify_func(identifier: str):
    """Identify a custom arithmetic algorithm with a unique name"""

    def wrapper(fn: Callable) -> CustomArithmeticAlgorithm:
        if identifier in _AI_REGISTRY.values():
            raise ValueError("Identifier already in use")
        if fn in _AI_REGISTRY:
            raise ValueError(
                f"Function already registered with identifier {_AI_REGISTRY[fn]}"
            )
        _AI_REGISTRY[fn] = identifier
        return fn

    return wrapper


def remove_func(func: CustomArithmeticAlgorithm):
    """Remove a custom arithmetic algorithm from the registry"""
    if func in _AI_REGISTRY:
        del _AI_REGISTRY[func]
        return True
    return False


def find_func(identifier: str):
    """Find the function of an identifier that has been registered"""
    return [k for k, v in _AI_REGISTRY.items() if v == identifier][0]


@dataclass
class PrecisionContext:
    """Context containing `PreciseNumber` arithmetic rules."""

    #: Precision mode
    pmode: PrecisionMode = PrecisionMode.MAX
    #: Truncature mode
    tmode: TruncatureMode = TruncatureMode.NONE
    #: Addition `CustomArithmeticAlgorithm`
    add: CustomArithmeticAlgorithm = None
    #: Substraction `CustomArithmeticAlgorithm`
    sub: CustomArithmeticAlgorithm = None
    #: Multiplication `CustomArithmeticAlgorithm`
    mul: CustomArithmeticAlgorithm = None
    #: Division `CustomArithmeticAlgorithm`
    div: CustomArithmeticAlgorithm = None
    #: Recording mode
    recording: bool = False

    #: `set_precision` context stack
    stack: int = field(init=False, default=0)
    _records: List = field(init=False, default_factory=list)

    def __post_init__(self):
        if not isinstance(self.tmode, TruncatureMode):
            raise TypeError

        if isinstance(self.pmode, int):
            if self.pmode < 0:
                raise ValueError("Precision cannot be negative")
            self._precisionfunc = lambda *_: self.pmode
        elif isinstance(self.pmode, PrecisionMode):
            self._precisionfunc = self.pmode
        else:
            raise TypeError

    def mutate(
        self,
        pmode: Optional[PrecisionMode] = None,
        tmode: Optional[TruncatureMode] = None,
        recording: Optional[bool] = None,
        add: Union[CustomArithmeticAlgorithm, Literal[False]] = False,
        sub: Union[CustomArithmeticAlgorithm, Literal[False]] = False,
        mul: Union[CustomArithmeticAlgorithm, Literal[False]] = False,
        div: Union[CustomArithmeticAlgorithm, Literal[False]] = False,
    ):
        """Mutates this `PrecisionContext` with new rules."""

        for f in [add, sub, mul, div]:
            if f is not False and f not in _AI_REGISTRY:
                raise ValueError("CustomArithmeticFunction not registered")

        self.pmode = self.pmode if pmode is None else pmode
        self.tmode = tmode or self.tmode
        self.recording = self.recording if recording is None else recording
        self.add = add if add is not False else self.add
        self.sub = sub if sub is not False else self.sub
        self.mul = mul if mul is not False else self.mul
        self.div = div if div is not False else self.div

        self.__post_init__()

    def freeze(self):
        """Returns a `Dict` containing this context rules"""
        return {
            "tmode": self.tmode.name,
            "pmode": self.pmode.name
            if isinstance(self.pmode, PrecisionMode)
            else self.pmode,
            "add": _AI_REGISTRY[self.add],
            "sub": _AI_REGISTRY[self.sub],
            "mul": _AI_REGISTRY[self.mul],
            "div": _AI_REGISTRY[self.div],
        }

    def record(self, *args):
        """Record an operation"""
        if self.recording:
            self._records.append({"args": args, **self.freeze()})


__CONTEXT = PrecisionContext()


def get_context() -> PrecisionContext:
    """Returns current context"""
    return __CONTEXT


def set_context(context: PrecisionContext):
    """Replace current `PrecisionContext`.

    :raises ValueError: Raise if you set a new context while inside a `set_precision` \
    context manager.
    """
    if get_context().stack > 0:
        raise ValueError("You can't change context while inside a precision_context")
    context.stack = 0
    global __CONTEXT
    __CONTEXT = context


def set_recording(flag: bool):
    """Set current `PrecisionContext` recording mode to `flag`.

    :raises ValueError: Raise if you set recording while inside a `set_precision` \
    context manager
    """
    ctx = get_context()
    if ctx.stack > 0:
        raise ValueError(
            "You can't start recording while inside a precision_context,\
            you should use recording=True instead"
        )
    ctx.recording = flag


def get_records():
    """Get current `PrecisionContext` records."""
    return get_context()._records


def clear_records():
    """Clear current `PrecisionContext` records."""
    get_context()._records.clear()


@contextmanager
def set_precision(
    pmode: Optional[PrecisionMode] = None,
    tmode: Optional[TruncatureMode] = None,
    recording: Optional[bool] = None,
    add: Union[CustomArithmeticAlgorithm, Literal[False]] = False,
    sub: Union[CustomArithmeticAlgorithm, Literal[False]] = False,
    mul: Union[CustomArithmeticAlgorithm, Literal[False]] = False,
    div: Union[CustomArithmeticAlgorithm, Literal[False]] = False,
):
    """Mutates the current `PrecisionContext` with the specified rules."""
    ctx = get_context()
    current = asdict(ctx)
    del current["_records"]
    del current["stack"]
    try:
        ctx.stack += 1
        ctx.mutate(pmode, tmode, recording, add, sub, mul, div)
        yield asdict(ctx)
    finally:
        ctx.mutate(**current)
        ctx.stack -= 1


class PrecisionError(Exception):
    pass



from .meta import TableType


class Sun(TableType):
    @classmethod
    def astro_id(cls):
        return 1

    equ_of_the_sun = 4
    equ_of_time = 1
    mean_motion_solar_sideral_long = 3
    mean_motion_solar_tropical_long = 2
    revolutions = 86
    solar_velocities = 20


class Mercury(TableType):
    @classmethod
    def astro_id(cls):
        return 2

    equ_anomaly_at_max_dist = 72
    equ_anomaly_at_mean_dist = 27
    equ_anomaly_at_min_dist = 73
    equ_center = 26
    equ_minuta_proportionalia = 91
    first_lat = 96
    lat_deviation = 100
    lat_double_arg = 101
    lat_incl = 98
    lat_slant = 99
    mean_motion_anomaly = 25
    planetary_stations = 30
    second_lat = 97
    total_equ_double_arg_table = 28
    true_position_double_arg = 92
    variable_hypotenuse = 90
    velocity_double_arg = 95
    velocity_eccentricity_component = 93
    velocity_epicycle_component = 94


class Venus(TableType):
    @classmethod
    def astro_id(cls):
        return 3

    equ_anomaly_at_max_dist = 76
    equ_anomaly_at_mean_dist = 33
    equ_anomaly_at_min_dist = 77
    equ_center = 32
    equ_minuta_proportionalia = 103
    first_lat = 108
    lat_deviation = 112
    lat_double_arg = 113
    lat_incl = 110
    lat_slant = 111
    mean_motion_anomaly = 31
    planetary_stations = 36
    second_lat = 109
    total_equ_double_arg_table = 34
    true_position_double_arg = 104
    variable_hypotenuse = 102
    velocity_double_arg = 107
    velocity_eccentricity_component = 105
    velocity_epicycle_component = 106


class Moon(TableType):
    @classmethod
    def astro_id(cls):
        return 4

    equ_anomaly = 47
    equ_anomaly_at_max_dist = 188
    equ_center = 46
    equ_minuta_proportionalia = 89
    equ_variation_of_diameter = 88
    lunar_lat = 48
    lunar_velocities = 23
    mean_arg_of_lat = 87
    mean_motion_double_elongation = 45
    mean_motion_long_plus_lunar_node = 43
    mean_motion_lunar_anomaly = 38
    mean_motion_lunar_elongation = 44
    mean_motion_lunar_long = 37
    mean_motion_lunar_node = 42


class Mars(TableType):
    @classmethod
    def astro_id(cls):
        return 5

    equ_anomaly_at_max_dist = 70
    equ_anomaly_at_mean_dist = 51
    equ_anomaly_at_min_dist = 71
    equ_center = 50
    equ_minuta_proportionalia = 115
    first_lat = 120
    lat_double_arg = 124
    mean_motion_long = 49
    northern_lat = 122
    planetary_stations = 54
    second_lat = 121
    southern_lat = 123
    total_equ_double_arg_table = 52
    true_position_double_arg = 116
    variable_hypotenuse = 114
    velocity_double_arg = 119
    velocity_eccentricity_component = 117
    velocity_epicycle_component = 118


class Jupiter(TableType):
    @classmethod
    def astro_id(cls):
        return 6

    equ_anomaly_at_max_dist = 68
    equ_anomaly_at_mean_dist = 57
    equ_anomaly_at_min_dist = 69
    equ_center = 56
    equ_minuta_proportionalia = 126
    first_lat = 131
    lat_double_arg = 135
    mean_motion_long = 55
    northern_lat = 133
    planetary_stations = 60
    second_lat = 132
    southern_lat = 134
    total_equ_double_arg_table = 58
    true_position_double_arg = 127
    variable_hypotenuse = 125
    velocity_double_arg = 130
    velocity_eccentricity_component = 128
    velocity_epicycle_component = 129


class Saturn(TableType):
    @classmethod
    def astro_id(cls):
        return 7

    equ_anomaly_at_max_dist = 74
    equ_anomaly_at_mean_dist = 63
    equ_anomaly_at_min_dist = 75
    equ_center = 62
    equ_minuta_proportionalia = 137
    first_lat = 142
    lat_double_arg = 146
    mean_motion_long = 61
    northern_lat = 144
    planetary_stations = 66
    second_lat = 143
    southern_lat = 145
    total_equ_double_arg_table = 64
    true_position_double_arg = 138
    variable_hypotenuse = 136
    velocity_double_arg = 141
    velocity_eccentricity_component = 139
    velocity_epicycle_component = 140


class SphericalAstronomical(TableType):
    @classmethod
    def astro_id(cls):
        return 8

    ascensional_diff = 11
    declination = 8
    houses = 79
    length_of_daylight = 78
    length_of_diurnal_seasonal_hour = 12
    meridian_altitude_of_the_sun = 80
    oblique_ascension = 10
    projection_of_rays = 83
    proportion_tables = 84
    right_ascension = 9
    solar_altitude_from_the_time_of_day = 82
    tangent = 85
    time_of_day_from_solar_altitude = 81


class EighthSphere(TableType):
    @classmethod
    def astro_id(cls):
        return 9

    equ_access_and_recess = 148
    mean_motion_access_and_recess = 13
    mean_motion_apogees_and_stars_precessions = 14
    mean_motion_solar_apogee = 17
    trepidation = 147


class Eclipse(TableType):
    @classmethod
    def astro_id(cls):
        return 10

    digit_of_lunar_eclipse_at_great_dist = 163
    digit_of_lunar_eclipse_at_near_dist = 166
    digit_of_solar_eclipse_at_great_dist = 169
    digit_of_solar_eclipse_at_near_dist = 171
    eclipse_interpolation_table = 173
    eclipsed_part_of_the_lunar_disk = 174
    eclipsed_part_of_the_solar_disk = 175
    half_totality_of_lunar_eclipse_at_great_dist = 165
    half_totality_of_lunar_eclipse_at_near_dist = 168
    immersion_of_eclipse_at_near_dist = 172
    immersion_of_lunar_eclipse_at_great_dist = 164
    immersion_of_lunar_eclipse_at_near_dist = 167
    immersion_of_solar_eclipse_at_great_dist = 170
    incl_at_the_beginning_of_lunar_eclipse = 178
    incl_at_the_beginning_of_solar_eclipse = 176
    incl_at_the_beginning_of_totality_for_a_lunar_eclipse = 179
    incl_at_the_end_of_lunar_eclipse = 181
    incl_at_the_end_of_solar_eclipse = 177
    incl_at_the_end_of_totality_for_a_lunar_eclipse = 180
    incl_local_lat_component = 185
    incl_obliquity_component = 186
    lunar_eclipse_half_duration_table = 183
    lunar_eclipse_half_totality_table = 184
    mean_motion_syzygies = 149
    moon_prlx_at_quadrature_and_true_anomaly_at_0deg = 154
    moon_prlx_at_quadrature_and_true_anomaly_at_180deg = 155
    moon_prlx_at_syzygy_and_true_anomaly_at_0deg = 152
    moon_prlx_at_syzygy_and_true_anomaly_at_180deg = 153
    moon_prlx_interpolation_at_quadrature = 157
    moon_prlx_interpolation_at_syzygy = 156
    moon_prlx_interpolation_between_syzygy_and_quadrature = 158
    prlx_lat_component = 151
    prlx_long_component = 150
    radius_of_the_moon = 160
    radius_of_the_shadow = 161
    radius_of_the_sun = 159
    solar_eclipse_half_duration_table = 182
    variation_of_shadow = 162


class Mathematical(TableType):
    @classmethod
    def astro_id(cls):
        return 11

    chords = 5
    cosine = 67
    shadow = 7
    sine = 6



import math as m

RAD = m.pi / 180
DEG = 1 / RAD


# # # # # # # # # # # #
#       PLANETS       #
# # # # # # # # # # # #


def planet_anomaly_0(x, R, rho):
    """
    anomaly of the planet or argument equation
    :param x: true argument in degree
    :param R: radius of epicycle
    :param rho distance between earth and epicycle center
    :return anomaly in degree
    """
    return DEG * m.atan(product_sine_0(x, R) / (rho + product_cosine_0(x, R)))


def q_2(x, e):
    """
    Planet center equation except Mercury,
     the function is preceded by a minus sign to agree with Pedersen
    :param x: mean center in degree
    :param e: eccentricity
    :return  center equation in degree
    """
    return -DEG * m.atan(
        2
        * product_sine_0(x, e)
        / (m.sqrt(60 ** 2 - (product_sine_0(x, e)) ** 2) + e * m.cos(x * RAD))
    )


# # # # # # # # # # # #
#      INFERIOR       #
#       PLANETS       #
# # # # # # # # # # # #


# JC FUNCTION
def lat_inf_slant_apo_def_0(x, b3m, pm, R):
    """
    inferior planet latitude at the deferent apogee owing to the slant of the epicycle but according
    with Ptolemy CALCULATION
    :param x: true argument in degree
    :param b3m: max of slant latitude  in degree
    :param pm: max of p(0º, av) in degree
    :param R: radius of the epicycle
    :return : part of the latitude  at the deferent apogee in degree
    """
    return planet_anomaly_0(x, R, 60) * b3m / pm


def lat_inf_slant_apo_def_1(x, km, rho, R):
    """
    inferior planet latitude at the deferent apogee owing to the slant of the epicycle but according
    with Ptolemy THEORY
    :param x: true argument in degree
    :param km: max of slant  in degree
    :param rho: distance earth-epicycle centre
    :param R: radius of the epicycle
    :return : part of the latitude  at the deferent apogee in degree
    """

    return DEG * m.asin(R * m.sin(RAD * km) * m.sin(RAD * x) / D_1(rho, x, R))


# JC FUNCTION
def lat_inf_devia_asc_node_0(av, jm, R):
    """
    inferior planet latitude at the ascending node owing to the deviation of the epicycle
    :param av: true argument in degree
    :jm: deviation max in degree
    :R: radius of the epicycle
    :return: part of the latitude at the ascending node in degree
    """

    return DEG * m.asin(R * m.sin(RAD * jm) * m.cos(RAD * av) / D_0(av, jm, R))


# JC FUNCTION
def lat_inf_incl_0(x, im):
    """
    part of inferior planets latitude owing
    to the inclination of the deferent
    :param x: nodal argument of latitude = angle between ascending node and epicycle center
    :im: inclination max in degree
    """
    return im * (m.sin(RAD * x)) ** 2


def minuta_proportionalia(x, R, e, cm):
    p0 = planet_anomaly_0(x, R, 60)
    p1 = planet_anomaly_0(x, R, 60 + e)
    p2 = planet_anomaly_0(x, R, 60 - e)
    maxp0 = max_equat_anomaly_0(R, 60)
    maxp1 = max_equat_anomaly_0(R, 60 + e)
    maxp2 = max_equat_anomaly_0(R, 60 - e)
    maxp = max_equat_anomaly_0(R, rho_1(cm, e))
    f1 = (maxp - maxp0) / (maxp1 - maxp0)
    f2 = (maxp - maxp0) / (maxp2 - maxp0)
    cm0 = mean_pos_epic_center_0(e)
    if cm > 180:
        cm = 360 - cm
    if 0 <= cm <= cm0:
        p = p0 - (p0 - p1) * f1
    else:
        p = p0 + (p2 - p0) * f2
    return p


# # # # # # # # # # # #
#      SUPERIOR       #
#       PLANETS       #
# # # # # # # # # # # #


# # # # # # # # # # # #
#      LATITUDE       #
#      LONGITUDE      #
# # # # # # # # # # # #

# # # # # # # # # # # #
#      SPHERICAL      #
#     ASTRONOMICAL    #
# # # # # # # # # # # #


def declin_0(x, obl):
    """
    declination of a point of ecliptic
    :param x: longitude of the point in degree
    :param obl: obliquity of the ecliptic in degree
    return declination in degree
    """
    return DEG * m.asin(m.sin(x * RAD) * m.sin(obl * RAD))


def right_asc_0(x, obl):
    """
    Right ascension of a point on ecliptic
    :param x: longitude in degree
    :param eps: inclination of ecliptic in degree
    :param inc: obliquity of the ecliptic
    return right ascension in degree
    define a continue function for x real.
    """
    n = m.floor(x / 180 + 1 / 2)
    ra = 0.0
    if x - (n - 1 / 2) * 180 > 0:
        ra = DEG * m.atan(m.cos(obl * RAD) * m.tan(x * RAD)) + n * 180

    if x == (n - 1 / 2) * 180:
        ra = (n - 1 / 2) * 180

    return ra


def oblique_asc_0(x, oblra, oblad, phi):
    """
    oblique ascension
    :param x: longitude in degree
    :param oblra: obliquity right ascension of the ecliptic in degree
    :param oblad: obliquity ascensional difference of the ecliptic in degree
    :param phi: latitude of observer in degree
    :return: oblique ascension in degree
    """
    return right_asc_0(x, oblra) - DEG * m.asin(
        m.tan(RAD * declin_0(x, oblad)) * m.tan(RAD * phi)
    )


# # # # # # # # # # # #
#         SUN         #
# # # # # # # # # # # #


# # # # # # # # # # # #
#   TRIGONOMETRICAL   #
# # # # # # # # # # # #


def product_sine_0(x, R):
    """
    :param x: angle in degree
    :param R: radius/eccentricity
    """
    return R * m.sin(x * RAD)


def product_cosine_0(x, R):
    """
    :param x: angle in degree
    :param R: factor
    """
    return R * m.cos(x * RAD)


def sin_double(x, y, param_0):
    """
    Sine with double argument
    :param x:
    :param y:
    :param param_0:
    :return:
    """
    return m.sin(RAD * x) * m.sin(RAD * y) * m.sin(RAD * param_0)


def chords_equation_0(x, param_0):
    """
    Chords equation
    :param x:
    :param param_0:
    :return:
    """
    R = param_0
    return 2 * R * m.sin(x * RAD / 2)


def shadow_equation_0(x, param_0):
    """
    Shadow table
    :param x:
    :param param_0:
    :return:
    """
    c = param_0
    return c / m.tan(x * RAD)


# # # # # # # # # # #
#      OTHERS       #
# # # # # # # # # # #


def D_0(av, jm, R):
    """
    Distance from earth to the inferior planet in relation to deviation in latitude
    :param av: true argument in degree
    :param jm: max deviation in degree
    :param R: radius of the epicycle
    """
    return m.sqrt(60 ** 2 + 120 * R * m.cos(RAD * jm) * m.cos(RAD * av) + R ** 2)


def D_1(rho, av, R):
    """
    Distance from earth to the inferior planet in relation to slant in latitude
    :param av: true argument in degree
    :param rho: distance earth-epicycle center
    :param R: radius of the epicycle
    """
    return m.sqrt(rho ** 2 + 2 * rho * R * m.cos(RAD * av) + R ** 2)


def q_0(x, e):
    """
    equation center of the sun  according to Pedersen
    x in degree, return in degree
    """
    return -DEG * m.atan(e * m.sin(RAD * x) / (60 + e * m.cos(RAD * x)))


def rho_0(cm, s, e):
    """
    Alkashi formula  with supplementary angle
    distance Earth-Epicycle centre  for mercury
    cm in degree
    """
    return m.sqrt(s ** 2 + e ** 2 + 2 * e * s * m.cos(RAD * cm))


def rho_1(cm, e):
    """
    distance earth-Epicycle center for venus and planet sup
    cm: mean center in degree
    e:eccentricity
    return: distance
    """
    base = m.sqrt(60 ** 2 - (e * m.sin(RAD * cm)) ** 2) + e * m.cos(RAD * cm)
    return m.sqrt(base ** 2 + (2 * e * m.sin(RAD * cm)) ** 2)


def rho_2(cm, e):
    """
    distance earth-Epicycle center for mercury
    cm: mean center in degree
    e:eccentricity
    return: distance
    """
    return rho_0(cm, s_m_0(cm, e), e)


def s_m_0(cm, e):
    """
    s fonction of mercury, distance equant-epicycle centre
    cm mean centre in degree
    e eccentricity
    """
    s1 = m.sqrt(60 ** 2 - e ** 2 * (m.sin(RAD * cm) + m.sin(RAD * 2 * cm)) ** 2)
    return s1 + e * (m.cos(RAD * cm) + m.cos(RAD * 2 * cm))


def mean_pos_epic_center_0(e):
    """
    mean center of the mean position of the epicycle center
    except Mercury, solution of the equation: rho(cm)=60
    e: eccentricity
    return: angle cm0 in degree
    """
    return 180 - DEG * m.atan(40 * m.sqrt(1 - (e / 120) ** 2) / e)


def mean_pos_epic_center_1(e):
    """
    Mean position of the epicycle centre in the case of mercury
    defined by mean centre cm. This value is called cm0.
    It is the approximate solution of the equation rho(cm)=60,
    with rho the distance between earth and epicycle centre
    return cm0 approximate at 1e-7 in degree
    """
    cm = 0
    eps = 1
    p = 1e-7
    while eps > p:
        while rho_2(cm, e) > 60:
            cm = cm + eps
        cm = cm - eps
        eps = eps / 10
    return cm


def max_equat_anomaly_0(r, rho):
    """
    Maximum of the equation anomaly
    rho:distance earth-epicycle centre
    r: radius of epicycle
    return: angle in degree
    """
    return DEG * m.asin(r / rho)


def q_1(cm, e):
    """
    Center equation of Mercury
    cm:mean centre;
    e:eccentricity
    """
    rho = rho_2(cm, e)
    return -DEG * m.asin(e * m.sin(RAD * cm) / rho)


def variation_of_centre_equation_0(cm, d, e):
    """
    variation_of_centre_equation for mercury
    cm:mean centre in degree
    d:step of the variation in degree
    e:eccentricity
    return variation number without dimension
    """
    return (q_1(cm + d, e) - q_1(cm, e)) / d


def variation_of_centre_equation_1(cm, d, e):
    """
    variation_of_centre_equation for venus and planet sup
    cm:mean centre in degree
    d:step of the variation in degree
    e:eccentricity
    return variation number without dimension
    """
    return (q_2(cm + d, e) - q_2(cm, e)) / d


def first_stationary_point_0(vq, dq, rho, r, e):
    """
    Give the true arg of first station for all planets
    vq:velocity quotient (wa/wt)
    dq:centre equation variation
    rho:distance earth-epicycle centre
    r:radius of epicycle
    e: eccentricity
    return :av(s1) = angle (DCZ) in degree=true argument of first station
    """
    A = (rho + r) * (rho - r)
    B = (1 + dq) / (vq - dq)
    X = m.sqrt(A / (1 + 2 * B))
    Y = 2 * B * X
    TCG = DEG * m.asin((2 * X + Y) / (2 * rho))
    TCZ = DEG * m.asin(Y / (2 * r))
    ZCG = TCG - TCZ
    return 180 - ZCG


def first_stationary_point_mercury_0(cm, vq, r, e):
    """
    True argument of first stationary point of  mercury by calculation in degree
    cm:mean centre in degree
    vq:velocity quotient (wa/wt)
    r:radius of epicycle
    e: eccentricity
    return :av(s1) = angle (DCZ) in degree for mercury
    """

    rho = rho_2(cm, e)
    dq = variation_of_centre_equation_0(cm, 6, e)
    return first_stationary_point_0(vq, dq, rho, r, e)


def first_stationary_point_all_planet_except_mercury_0(cm, vq, r, e):
    """
    True argument of first stationary point of  all planet except mercury by calculation in degree
    cm:mean centre in degree
    vq:velocity quotient (wa/wt)
    r:radius of epicycle
    e: eccentricity
    return :av(s1) = angle (DCZ) in degree
    """

    rho = rho_1(cm, e)
    dq = variation_of_centre_equation_1(cm, 6, e)
    return first_stationary_point_0(vq, dq, rho, r, e)


def first_stationary_point_proport_minutes_0(x, e, s0, s1, s2):
    """
    True argument of first stationary point of sup planet or venus (except mercury)
    by proportional minutes in degree
    :param x: mean centre in degree (0<=x<360)
    :param e: planetary eccentricity
    :param s0: first station. true arg. at the apogee (cm=0)
    :param s1: first station. true arg. at the mean position (cm=cm0)
    :param s2: first station. true arg. at the perigee (cm=180)
    :return:true argument of the planet first station in degree
    """

    cm0 = mean_pos_epic_center_0(e)
    R0 = rho_1(0, e)
    R1 = rho_1(cm0, e)
    R2 = rho_1(180, e)
    R = rho_1(x, e)

    cm = x
    if (180 < x) and (x < 360):
        cm = 360 - x
    if (0 <= cm) and (cm < cm0):
        s = s0 + (s1 - s0) * (R - R0) / (R1 - R0)
    if (cm0 <= cm) and (cm <= 180):
        s = s1 + (s2 - s1) * (R - R1) / (R2 - R1)
    return s


def zero_equation_0(x, param_0):
    """
    Zero equation
    :param x:
    :param param_0:
    :return:
    """
    return 0


def zero_double_equation_0(x, y, param_0):
    """
    Zero double argument equation
    :param x:
    :param y:
    :param param_0:
    :return:
    """
    return 0


def prod_double(x, y, param_0):
    """
    Product
    :param x:
    :param y:
    :param param_0:
    :return:
    """
    return x * y * param_0


def sum_double(x, y, param_0):
    """
    Asymmetric sum
    :param x:
    :param y:
    :param param_0:
    :return:
    """
    return (x + y ** 2) * param_0



import pytest

from kanon.units import RadixBase, radix_registry


class TestRadixBase:
    def test_bases(self):
        test_radix = RadixBase([1], [2], "Test", "a")

        assert "Test" in radix_registry
        assert radix_registry["Test"].base is test_radix

        assert test_radix.mixed

        with pytest.raises(ValueError):
            RadixBase([1], [2], "Sexagesimal")

        with pytest.raises(ValueError):
            RadixBase([11, 5], [2], "a", integer_separators=["a", "b", "c"])

        with pytest.raises(AssertionError):
            RadixBase([], [2], "n1")
        with pytest.raises(AssertionError):
            RadixBase([1], [1.1], "n1")

    def test_get(self):

        sexa_base = radix_registry["Sexagesimal"].base

        assert sexa_base[1] == 60
        assert sexa_base[-1] == 60
        assert sexa_base[100] == 60

        histo_base = radix_registry["Historical"].base
        assert histo_base[-2:1] == [10, 12, 30]

        with pytest.raises(ValueError):
            histo_base[5:]
        with pytest.raises(ValueError):
            histo_base[:3]
